# Copyright (C) 2007-2011 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the BranchBuilder class."""

from .. import branch as _mod_branch
from .. import revision as _mod_revision
from .. import tests
from ..branchbuilder import BranchBuilder
from ..bzr import branch as _mod_bzrbranch


class TestBranchBuilder(tests.TestCaseWithMemoryTransport):
    def test_create(self):
        """Test the constructor api."""
        BranchBuilder(self.get_transport().clone("foo"))
        # we dont care if the branch has been built or not at this point.

    def test_get_branch(self):
        """get_branch returns the created branch."""
        builder = BranchBuilder(self.get_transport().clone("foo"))
        branch = builder.get_branch()
        self.assertIsInstance(branch, _mod_branch.Branch)
        self.assertEqual(self.get_transport().clone("foo").base, branch.base)
        self.assertEqual((0, _mod_revision.NULL_REVISION), branch.last_revision_info())

    def test_format(self):
        """Making a BranchBuilder with a format option sets the branch type."""
        builder = BranchBuilder(self.get_transport(), format="dirstate-tags")
        branch = builder.get_branch()
        self.assertIsInstance(branch, _mod_bzrbranch.BzrBranch6)

    def test_build_one_commit(self):
        """Doing build_commit causes a commit to happen."""
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id = builder.build_commit()
        branch = builder.get_branch()
        self.assertEqual((1, rev_id), branch.last_revision_info())
        self.assertEqual(
            "commit 1", branch.repository.get_revision(branch.last_revision()).message
        )

    def test_build_commit_timestamp(self):
        """You can set a date when committing."""
        builder = self.make_branch_builder("foo")
        rev_id = builder.build_commit(timestamp=1236043340)
        branch = builder.get_branch()
        self.assertEqual((1, rev_id), branch.last_revision_info())
        rev = branch.repository.get_revision(branch.last_revision())
        self.assertEqual("commit 1", rev.message)
        self.assertEqual(1236043340, int(rev.timestamp))

    def test_build_two_commits(self):
        """The second commit has the right parents and message."""
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id1 = builder.build_commit()
        rev_id2 = builder.build_commit()
        branch = builder.get_branch()
        self.assertEqual((2, rev_id2), branch.last_revision_info())
        self.assertEqual(
            "commit 2", branch.repository.get_revision(branch.last_revision()).message
        )
        self.assertEqual(
            [rev_id1], branch.repository.get_revision(branch.last_revision()).parent_ids
        )

    def test_build_commit_parent_ids(self):
        """build_commit() takes a parent_ids argument."""
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id1 = builder.build_commit(
            parent_ids=[b"ghost"], allow_leftmost_as_ghost=True
        )
        rev_id2 = builder.build_commit(parent_ids=[])
        branch = builder.get_branch()
        self.assertEqual((1, rev_id2), branch.last_revision_info())
        self.assertEqual([b"ghost"], branch.repository.get_revision(rev_id1).parent_ids)


class TestBranchBuilderBuildSnapshot(tests.TestCaseWithMemoryTransport):
    def assertTreeShape(self, expected_shape, tree):
        """Check that the tree shape matches expectations."""
        tree.lock_read()
        try:
            entries = [
                (path, ie.file_id, ie.kind) for path, ie in tree.iter_entries_by_dir()
            ]
        finally:
            tree.unlock()
        self.assertEqual(expected_shape, entries)

    def build_a_rev(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id1 = builder.build_snapshot(
            None,
            [
                ("add", ("", b"a-root-id", "directory", None)),
                ("add", ("a", b"a-id", "file", b"contents")),
            ],
            revision_id=b"A-id",
        )
        self.assertEqual(b"A-id", rev_id1)
        return builder

    def test_add_one_file(self):
        builder = self.build_a_rev()
        branch = builder.get_branch()
        self.assertEqual((1, b"A-id"), branch.last_revision_info())
        rev_tree = branch.repository.revision_tree(b"A-id")
        rev_tree.lock_read()
        self.addCleanup(rev_tree.unlock)
        self.assertTreeShape(
            [("", b"a-root-id", "directory"), ("a", b"a-id", "file")], rev_tree
        )
        self.assertEqual(b"contents", rev_tree.get_file_text("a"))

    def test_add_second_file(self):
        builder = self.build_a_rev()
        rev_id2 = builder.build_snapshot(
            None, [("add", ("b", b"b-id", "file", b"content_b"))], revision_id=b"B-id"
        )
        self.assertEqual(b"B-id", rev_id2)
        branch = builder.get_branch()
        self.assertEqual((2, rev_id2), branch.last_revision_info())
        rev_tree = branch.repository.revision_tree(rev_id2)
        rev_tree.lock_read()
        self.addCleanup(rev_tree.unlock)
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("b", b"b-id", "file"),
            ],
            rev_tree,
        )
        self.assertEqual(b"content_b", rev_tree.get_file_text("b"))

    def test_add_empty_dir(self):
        builder = self.build_a_rev()
        builder.build_snapshot(
            None, [("add", ("b", b"b-id", "directory", None))], revision_id=b"B-id"
        )
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("b", b"b-id", "directory"),
            ],
            rev_tree,
        )

    def test_commit_timestamp(self):
        builder = self.make_branch_builder("foo")
        rev_id = builder.build_snapshot(
            None, [("add", ("", None, "directory", None))], timestamp=1234567890
        )
        rev = builder.get_branch().repository.get_revision(rev_id)
        self.assertEqual(1234567890, int(rev.timestamp))

    def test_commit_message_default(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id = builder.build_snapshot(None, [("add", ("", None, "directory", None))])
        branch = builder.get_branch()
        rev = branch.repository.get_revision(rev_id)
        self.assertEqual("commit 1", rev.message)

    def test_commit_message_supplied(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id = builder.build_snapshot(
            None, [("add", ("", None, "directory", None))], message="Foo"
        )
        branch = builder.get_branch()
        rev = branch.repository.get_revision(rev_id)
        self.assertEqual("Foo", rev.message)

    def test_commit_message_callback(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        rev_id = builder.build_snapshot(
            None,
            [("add", ("", None, "directory", None))],
            message_callback=lambda x: "Foo",
        )
        branch = builder.get_branch()
        rev = branch.repository.get_revision(rev_id)
        self.assertEqual("Foo", rev.message)

    def test_modify_file(self):
        builder = self.build_a_rev()
        rev_id2 = builder.build_snapshot(
            None, [("modify", ("a", b"new\ncontent\n"))], revision_id=b"B-id"
        )
        self.assertEqual(b"B-id", rev_id2)
        branch = builder.get_branch()
        rev_tree = branch.repository.revision_tree(rev_id2)
        rev_tree.lock_read()
        self.addCleanup(rev_tree.unlock)
        self.assertEqual(b"new\ncontent\n", rev_tree.get_file_text("a"))

    def test_delete_file(self):
        builder = self.build_a_rev()
        rev_id2 = builder.build_snapshot(
            None, [("unversion", "a")], revision_id=b"B-id"
        )
        self.assertEqual(b"B-id", rev_id2)
        branch = builder.get_branch()
        rev_tree = branch.repository.revision_tree(rev_id2)
        rev_tree.lock_read()
        self.addCleanup(rev_tree.unlock)
        self.assertTreeShape([("", b"a-root-id", "directory")], rev_tree)

    def test_delete_directory(self):
        builder = self.build_a_rev()
        builder.build_snapshot(
            None,
            [
                ("add", ("b", b"b-id", "directory", None)),
                ("add", ("b/c", b"c-id", "file", b"foo\n")),
                ("add", ("b/d", b"d-id", "directory", None)),
                ("add", ("b/d/e", b"e-id", "file", b"eff\n")),
            ],
            revision_id=b"B-id",
        )
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("b", b"b-id", "directory"),
                ("b/c", b"c-id", "file"),
                ("b/d", b"d-id", "directory"),
                ("b/d/e", b"e-id", "file"),
            ],
            rev_tree,
        )
        # Removing a directory removes all child dirs
        builder.build_snapshot(None, [("unversion", "b")], revision_id=b"C-id")
        rev_tree = builder.get_branch().repository.revision_tree(b"C-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
            ],
            rev_tree,
        )

    def test_unknown_action(self):
        builder = self.build_a_rev()
        e = self.assertRaises(
            ValueError,
            builder.build_snapshot,
            None,
            [("weirdo", ("foo",))],
            revision_id=b"B-id",
        )
        self.assertEqual('Unknown build action: "weirdo"', str(e))

    def test_rename(self):
        builder = self.build_a_rev()
        builder.build_snapshot(None, [("rename", ("a", "b"))], revision_id=b"B-id")
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape(
            [("", b"a-root-id", "directory"), ("b", b"a-id", "file")], rev_tree
        )

    def test_rename_into_subdir(self):
        builder = self.build_a_rev()
        builder.build_snapshot(
            None,
            [
                ("add", ("dir", b"dir-id", "directory", None)),
                ("rename", ("a", "dir/a")),
            ],
            revision_id=b"B-id",
        )
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("dir", b"dir-id", "directory"),
                ("dir/a", b"a-id", "file"),
            ],
            rev_tree,
        )

    def test_rename_out_of_unversioned_subdir(self):
        builder = self.build_a_rev()
        builder.build_snapshot(
            None,
            [
                ("add", ("dir", b"dir-id", "directory", None)),
                ("rename", ("a", "dir/a")),
            ],
            revision_id=b"B-id",
        )
        builder.build_snapshot(
            None,
            [("rename", ("dir/a", "a")), ("unversion", "dir")],
            revision_id=b"C-id",
        )
        rev_tree = builder.get_branch().repository.revision_tree(b"C-id")
        self.assertTreeShape(
            [("", b"a-root-id", "directory"), ("a", b"a-id", "file")], rev_tree
        )

    def test_set_parent(self):
        builder = self.build_a_rev()
        builder.start_series()
        self.addCleanup(builder.finish_series)
        builder.build_snapshot(
            [b"A-id"], [("modify", ("a", b"new\ncontent\n"))], revision_id=b"B-id"
        )
        builder.build_snapshot(
            [b"A-id"],
            [("add", ("c", b"c-id", "file", b"alt\ncontent\n"))],
            revision_id=b"C-id",
        )
        # We should now have a graph:
        #   A
        #   |\
        #   C B
        # And not A => B => C
        repo = builder.get_branch().repository
        self.assertEqual(
            {b"B-id": (b"A-id",), b"C-id": (b"A-id",)},
            repo.get_parent_map([b"B-id", b"C-id"]),
        )
        b_tree = repo.revision_tree(b"B-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
            ],
            b_tree,
        )
        self.assertEqual(b"new\ncontent\n", b_tree.get_file_text("a"))

        # We should still be using the content from A in C, not from B
        c_tree = repo.revision_tree(b"C-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("c", b"c-id", "file"),
            ],
            c_tree,
        )
        self.assertEqual(b"contents", c_tree.get_file_text("a"))
        self.assertEqual(b"alt\ncontent\n", c_tree.get_file_text("c"))

    def test_set_merge_parent(self):
        builder = self.build_a_rev()
        builder.start_series()
        self.addCleanup(builder.finish_series)
        builder.build_snapshot(
            [b"A-id"],
            [("add", ("b", b"b-id", "file", b"b\ncontent\n"))],
            revision_id=b"B-id",
        )
        builder.build_snapshot(
            [b"A-id"],
            [("add", ("c", b"c-id", "file", b"alt\ncontent\n"))],
            revision_id=b"C-id",
        )
        builder.build_snapshot([b"B-id", b"C-id"], [], revision_id=b"D-id")
        repo = builder.get_branch().repository
        self.assertEqual(
            {b"B-id": (b"A-id",), b"C-id": (b"A-id",), b"D-id": (b"B-id", b"C-id")},
            repo.get_parent_map([b"B-id", b"C-id", b"D-id"]),
        )
        d_tree = repo.revision_tree(b"D-id")
        # Note: by default a merge node does *not* pull in the changes from the
        #       merged tree, you have to supply it yourself.
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("b", b"b-id", "file"),
            ],
            d_tree,
        )

    def test_set_merge_parent_and_contents(self):
        builder = self.build_a_rev()
        builder.start_series()
        self.addCleanup(builder.finish_series)
        builder.build_snapshot(
            [b"A-id"],
            [("add", ("b", b"b-id", "file", b"b\ncontent\n"))],
            revision_id=b"B-id",
        )
        builder.build_snapshot(
            [b"A-id"],
            [("add", ("c", b"c-id", "file", b"alt\ncontent\n"))],
            revision_id=b"C-id",
        )
        builder.build_snapshot(
            [b"B-id", b"C-id"],
            [("add", ("c", b"c-id", "file", b"alt\ncontent\n"))],
            revision_id=b"D-id",
        )
        repo = builder.get_branch().repository
        self.assertEqual(
            {b"B-id": (b"A-id",), b"C-id": (b"A-id",), b"D-id": (b"B-id", b"C-id")},
            repo.get_parent_map([b"B-id", b"C-id", b"D-id"]),
        )
        d_tree = repo.revision_tree(b"D-id")
        self.assertTreeShape(
            [
                ("", b"a-root-id", "directory"),
                ("a", b"a-id", "file"),
                ("b", b"b-id", "file"),
                ("c", b"c-id", "file"),
            ],
            d_tree,
        )
        # Because we copied the exact text into *this* tree, the 'c' file
        # should look like it was not modified in the merge
        self.assertEqual(b"C-id", d_tree.get_file_revision("c"))

    def test_set_parent_to_null(self):
        builder = self.build_a_rev()
        builder.start_series()
        self.addCleanup(builder.finish_series)
        builder.build_snapshot(
            [], [("add", ("", None, "directory", None))], revision_id=b"B-id"
        )
        # We should now have a graph:
        #   A B
        # And not A => B
        repo = builder.get_branch().repository
        self.assertEqual(
            {
                b"A-id": (_mod_revision.NULL_REVISION,),
                b"B-id": (_mod_revision.NULL_REVISION,),
            },
            repo.get_parent_map([b"A-id", b"B-id"]),
        )

    def test_start_finish_series(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        try:
            self.assertIsNot(None, builder._tree)
            self.assertEqual("w", builder._tree._lock_mode)
            self.assertTrue(builder._branch.is_locked())
        finally:
            builder.finish_series()
        self.assertIs(None, builder._tree)
        self.assertFalse(builder._branch.is_locked())

    def test_ghost_mainline_history(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        try:
            builder.build_snapshot(
                [b"ghost"],
                [("add", ("", b"ROOT_ID", "directory", ""))],
                allow_leftmost_as_ghost=True,
                revision_id=b"tip",
            )
        finally:
            builder.finish_series()
        b = builder.get_branch()
        b.lock_read()
        self.addCleanup(b.unlock)
        self.assertEqual(
            (b"ghost",), b.repository.get_graph().get_parent_map([b"tip"])[b"tip"]
        )

    def test_unversion_root_add_new_root(self):
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        builder.build_snapshot(
            None, [("add", ("", b"TREE_ROOT", "directory", ""))], revision_id=b"rev-1"
        )
        builder.build_snapshot(
            None,
            [("unversion", ""), ("add", ("", b"my-root", "directory", ""))],
            revision_id=b"rev-2",
        )
        builder.finish_series()
        rev_tree = builder.get_branch().repository.revision_tree(b"rev-2")
        self.assertTreeShape([("", b"my-root", "directory")], rev_tree)

    def test_empty_flush(self):
        """A flush with no actions before it is a no-op."""
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        builder.build_snapshot(
            None, [("add", ("", b"TREE_ROOT", "directory", ""))], revision_id=b"rev-1"
        )
        builder.build_snapshot(None, [("flush", None)], revision_id=b"rev-2")
        builder.finish_series()
        rev_tree = builder.get_branch().repository.revision_tree(b"rev-2")
        self.assertTreeShape([("", b"TREE_ROOT", "directory")], rev_tree)

    def test_kind_change(self):
        """It's possible to change the kind of an entry in a single snapshot
        with a bit of help from the 'flush' action.
        """
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"a-root-id", "directory", None)),
                ("add", ("a", b"a-id", "file", b"content\n")),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot(
            None,
            [
                ("unversion", "a"),
                ("flush", None),
                ("add", ("a", b"a-id", "directory", None)),
            ],
            revision_id=b"B-id",
        )
        builder.finish_series()
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape(
            [("", b"a-root-id", "directory"), ("a", b"a-id", "directory")], rev_tree
        )

    def test_pivot_root(self):
        """It's possible (albeit awkward) to move an existing dir to the root
        in a single snapshot by using unversion then flush then add.
        """
        builder = BranchBuilder(self.get_transport().clone("foo"))
        builder.start_series()
        builder.build_snapshot(
            None,
            [
                ("add", ("", b"orig-root", "directory", None)),
                ("add", ("dir", b"dir-id", "directory", None)),
            ],
            revision_id=b"A-id",
        )
        builder.build_snapshot(
            None,
            [
                ("unversion", ""),  # implicitly unversions all children
                ("flush", None),
                ("add", ("", b"dir-id", "directory", None)),
            ],
            revision_id=b"B-id",
        )
        builder.finish_series()
        rev_tree = builder.get_branch().repository.revision_tree(b"B-id")
        self.assertTreeShape([("", b"dir-id", "directory")], rev_tree)



# Copyright (C) 2006, 2008, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
#
# cython: language_level=3

"""Wrapper for readdir which returns files ordered by inode."""


import os
import sys


cdef extern from "python-compat.h":
    pass


cdef extern from 'errno.h':
    int ENOENT
    int ENOTDIR
    int EAGAIN
    int EINTR
    char *strerror(int errno)
    # not necessarily a real variable, but this should be close enough
    int errno

cdef extern from 'unistd.h':
    int chdir(char *path)
    int close(int fd)
    int fchdir(int fd)
    char *getcwd(char *, int size)

cdef extern from 'stdlib.h':
    void *malloc(int)
    void free(void *)


cdef extern from 'sys/types.h':
    ctypedef long ssize_t
    ctypedef unsigned long size_t
    ctypedef long time_t
    ctypedef unsigned long ino_t
    ctypedef unsigned long long off_t
    ctypedef int mode_t


cdef extern from 'sys/stat.h':
    cdef struct stat:
        int st_mode
        off_t st_size
        int st_dev
        ino_t st_ino
        int st_mtime
        int st_ctime
    int lstat(char *path, stat *buf)
    int S_ISDIR(int mode)
    int S_ISCHR(int mode)
    int S_ISBLK(int mode)
    int S_ISREG(int mode)
    int S_ISFIFO(int mode)
    int S_ISLNK(int mode)
    int S_ISSOCK(int mode)


cdef extern from 'fcntl.h':
    int O_RDONLY
    int open(char *pathname, int flags, mode_t mode)


cdef extern from 'Python.h':
    int PyErr_CheckSignals() except -1
    char * PyBytes_AS_STRING(object)
    ctypedef struct PyObject:
        pass
    Py_ssize_t PyBytes_Size(object s)
    object PyList_GetItem(object lst, Py_ssize_t index)
    void *PyList_GetItem_object_void "PyList_GET_ITEM" (object lst, int index)
    int PyList_Append(object lst, object item) except -1
    void *PyTuple_GetItem_void_void "PyTuple_GET_ITEM" (void* tpl, int index)
    int PyTuple_SetItem(void *, Py_ssize_t pos, object item) except -1
    int PyTuple_SetItem_obj "PyTuple_SetItem" (void *, Py_ssize_t pos, PyObject * item) except -1
    void Py_INCREF(object o)
    void Py_DECREF(object o)
    void PyBytes_Concat(PyObject **string, object newpart)


cdef extern from 'dirent.h':
    ctypedef struct dirent:
        char d_name[256]
        ino_t d_ino
    # the opaque C library DIR type.
    ctypedef struct DIR
    # should be DIR *, pyrex barfs.
    DIR * opendir(char * name)
    int closedir(DIR * dir)
    dirent *readdir(DIR *dir)

cdef object _directory
_directory = 'directory'
cdef object _chardev
_chardev = 'chardev'
cdef object _block
_block = 'block'
cdef object _file
_file = 'file'
cdef object _fifo
_fifo = 'fifo'
cdef object _symlink
_symlink = 'symlink'
cdef object _socket
_socket = 'socket'
cdef object _unknown
_unknown = 'unknown'

# add a typedef struct dirent dirent to workaround pyrex
cdef extern from 'readdir.h':
    pass


cdef class _Stat:
    """Represent a 'stat' result."""

    cdef stat _st

    property st_dev:
        def __get__(self):
            return self._st.st_dev

    property st_ino:
        def __get__(self):
            return self._st.st_ino

    property st_mode:
        def __get__(self):
            return self._st.st_mode

    property st_ctime:
        def __get__(self):
            return self._st.st_ctime

    property st_mtime:
        def __get__(self):
            return self._st.st_mtime

    property st_size:
        def __get__(self):
            return self._st.st_size

    def __repr__(self):
        """Repr is the same as a Stat object.

        (mode, ino, dev, nlink, uid, gid, size, None(atime), mtime, ctime)
        """
        return repr((self.st_mode, 0, 0, 0, 0, 0, self.st_size, None,
                     self.st_mtime, self.st_ctime))


from . import osutils


cdef object _safe_utf8
_safe_utf8 = osutils.safe_utf8

cdef class UTF8DirReader:
    """A dir reader for utf8 file systems."""

    def kind_from_mode(self, int mode):
        """Get the kind of a path from a mode status."""
        return self._kind_from_mode(mode)

    cdef _kind_from_mode(self, int mode):
        # Files and directories are the most common - check them first.
        if S_ISREG(mode):
            return _file
        if S_ISDIR(mode):
            return _directory
        if S_ISCHR(mode):
            return _chardev
        if S_ISBLK(mode):
            return _block
        if S_ISLNK(mode):
            return _symlink
        if S_ISFIFO(mode):
            return _fifo
        if S_ISSOCK(mode):
            return _socket
        return _unknown

    def top_prefix_to_starting_dir(self, top, prefix=""):
        """See DirReader.top_prefix_to_starting_dir."""
        return (_safe_utf8(prefix), None, None, None, _safe_utf8(top))

    def read_dir(self, prefix, top):
        """Read a single directory from a utf8 file system.

        All paths in and out are utf8.

        This sub-function is called when we know the filesystem is already in utf8
        encoding. So we don't need to transcode filenames.

        See DirReader.read_dir for details.
        """
        #cdef char *_prefix = prefix
        #cdef char *_top = top
        # Use C accelerated directory listing.
        cdef object newval
        cdef int index
        cdef int length
        cdef void * atuple
        cdef object name
        cdef PyObject * new_val_obj

        if PyBytes_Size(prefix):
            relprefix = prefix + b'/'
        else:
            relprefix = b''
        top_slash = top + b'/'

        # read_dir supplies in should-stat order.
        # for _, name in sorted(_listdir(top)):
        result = _read_dir(top)
        length = len(result)
        # result.sort()
        for index from 0 <= index < length:
            atuple = PyList_GetItem_object_void(result, index)
            name = <object>PyTuple_GetItem_void_void(atuple, 1)
            # We have a tuple with (inode, name, None, statvalue, None)
            # Now edit it:
            # inode -> path_from_top
            # direct concat - faster than operator +.
            new_val_obj = <PyObject *>relprefix
            Py_INCREF(relprefix)
            PyBytes_Concat(&new_val_obj, name)
            if NULL == new_val_obj:
                # PyBytes_Concat will have setup an exception, but how to get
                # at it?
                raise Exception("failed to strcat")
            PyTuple_SetItem_obj(atuple, 0, new_val_obj)
            # 1st None -> kind
            newval = self._kind_from_mode(
                (<_Stat>PyTuple_GetItem_void_void(atuple, 3)).st_mode)
            Py_INCREF(newval)
            PyTuple_SetItem(atuple, 2, newval)
            # 2nd None -> abspath # for all - the caller may need to stat files
            # etc.
            # direct concat - faster than operator +.
            new_val_obj = <PyObject *>top_slash
            Py_INCREF(top_slash)
            PyBytes_Concat(&new_val_obj, name)
            if NULL == new_val_obj:
                # PyBytes_Concat will have setup an exception, but how to get
                # at it?
                raise Exception("failed to strcat")
            PyTuple_SetItem_obj(atuple, 4, new_val_obj)
        return result


cdef raise_os_error(int errnum, char *msg_prefix, path):
    if errnum == EINTR:
        PyErr_CheckSignals()
    raise OSError(errnum, msg_prefix + strerror(errnum), path)


cdef _read_dir(path):
    """Like os.listdir, this reads the contents of a directory.

    :param path: the directory to list.
    :return: a list of single-owner (the list) tuples ready for editing into
        the result tuples walkdirs needs to yield. They contain (inode, name,
        None, statvalue, None).
    """
    cdef DIR *the_dir
    # currently this needs a fixup - the C code says 'dirent' but should say
    # 'struct dirent'
    cdef dirent * entry
    cdef dirent sentinel
    cdef char *name
    cdef int stat_result
    cdef _Stat statvalue
    global errno
    cdef int orig_dir_fd

    # Avoid chdir('') because it causes problems on Sun OS, and avoid this if
    # staying in .
    if path != b"" and path != b'.':
        # we change into the requested directory before reading, and back at the
        # end, because that turns out to make the stat calls measurably faster than
        # passing full paths every time.
        orig_dir_fd = open(".", O_RDONLY, 0)
        if orig_dir_fd == -1:
            raise_os_error(errno, "open: ", ".")
        if -1 == chdir(path):
            # Ignore the return value, because we are already raising an
            # exception
            close(orig_dir_fd)
            raise_os_error(errno, "chdir: ", path)
    else:
        orig_dir_fd = -1

    try:
        the_dir = opendir(b".")
        if NULL == the_dir:
            raise_os_error(errno, "opendir: ", path)
        try:
            result = []
            entry = &sentinel
            while entry != NULL:
                # Unlike most libc functions, readdir needs errno set to 0
                # beforehand so that eof can be distinguished from errors.  See
                # <https://bugs.launchpad.net/bzr/+bug/279381>
                while True:
                    errno = 0
                    entry = readdir(the_dir)
                    if entry == NULL and (errno == EAGAIN or errno == EINTR):
                        if errno == EINTR:
                            PyErr_CheckSignals()
                        # try again
                        continue
                    else:
                        break
                if entry == NULL:
                    if errno == ENOTDIR or errno == 0:
                        # We see ENOTDIR at the end of a normal directory.
                        # As ENOTDIR for read_dir(file) is triggered on opendir,
                        # we consider ENOTDIR to be 'no error'.
                        continue
                    else:
                        raise_os_error(errno, "readdir: ", path)
                name = entry.d_name
                if not (name[0] == c"." and (
                    (name[1] == 0) or 
                    (name[1] == c"." and name[2] == 0))
                    ):
                    statvalue = _Stat()
                    stat_result = lstat(entry.d_name, &statvalue._st)
                    if stat_result != 0:
                        if errno != ENOENT:
                            raise_os_error(errno, "lstat: ",
                                path + b"/" + entry.d_name)
                        else:
                            # the file seems to have disappeared after being
                            # seen by readdir - perhaps a transient temporary
                            # file.  there's no point returning it.
                            continue
                    # We append a 5-tuple that can be modified in-place by the C
                    # api:
                    # inode to sort on (to replace with top_path)
                    # name (to keep)
                    # kind (None, to set)
                    # statvalue (to keep)
                    # abspath (None, to set)
                    PyList_Append(result, (entry.d_ino, entry.d_name, None,
                        statvalue, None))
        finally:
            if -1 == closedir(the_dir):
                raise_os_error(errno, "closedir: ", path)
    finally:
        if -1 != orig_dir_fd:
            failed = False
            if -1 == fchdir(orig_dir_fd):
                # try to close the original directory anyhow
                failed = True
            if -1 == close(orig_dir_fd) or failed:
                raise_os_error(errno, "return to orig_dir: ", "")

    return result


# vim: tw=79 ai expandtab sw=4 sts=4



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test that all Tree's implement get_symlink_target."""

import os

from breezy.tests import features, per_tree


class TestGetSymlinkTarget(per_tree.TestCaseWithTree):
    def get_tree_with_symlinks(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        tree = self.make_branch_and_tree("tree")
        os.symlink("foo", "tree/link")
        os.symlink("../bar", "tree/rel_link")
        os.symlink("/baz/bing", "tree/abs_link")

        tree.add(["link", "rel_link", "abs_link"])
        return self._convert_tree(tree)

    def test_get_symlink_target(self):
        tree = self.get_tree_with_symlinks()
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual("foo", tree.get_symlink_target("link"))
        self.assertEqual("../bar", tree.get_symlink_target("rel_link"))
        self.assertEqual("/baz/bing", tree.get_symlink_target("abs_link"))
        self.assertEqual("foo", tree.get_symlink_target("link"))

    def test_get_unicode_symlink_target(self):
        self.requireFeature(features.SymlinkFeature(self.test_dir))
        self.requireFeature(features.UnicodeFilenameFeature)
        tree = self.make_branch_and_tree("tree")
        target = "targ\N{EURO SIGN}t"
        os.symlink(target, os.fsencode("tree/\u03b2_link"))
        tree.add(["\u03b2_link"])
        tree.lock_read()
        self.addCleanup(tree.unlock)
        actual = tree.get_symlink_target("\u03b2_link")
        self.assertEqual(target, actual)



# Copyright (C) 2005-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import time

from breezy import errors
from breezy import revision as _mod_revision
from breezy.revisionspec import (
    InvalidRevisionSpec,
    RevisionInfo,
    RevisionSpec,
    RevisionSpec_dwim,
    RevisionSpec_tag,
)
from breezy.tests import TestCaseWithTransport


def spec_in_history(spec, branch):
    """A simple helper to change a revision spec into a branch search."""
    return RevisionSpec.from_string(spec).in_history(branch)


# Basic class, which just creates a really basic set of revisions
class TestRevisionSpec(TestCaseWithTransport):
    def setUp(self):
        super().setUp()
        # this sets up a revision graph:
        # r1: []             1
        # alt_r2: [r1]       1.1.1
        # r2: [r1, alt_r2]   2

        self.tree = self.make_branch_and_tree("tree")
        self.build_tree(["tree/a"])
        self.tree.lock_write()
        self.addCleanup(self.tree.unlock)
        self.tree.add(["a"])
        self.tree.commit("a", rev_id=b"r1")

        self.tree2 = self.tree.controldir.sprout("tree2").open_workingtree()
        self.tree2.commit("alt", rev_id=b"alt_r2")

        self.tree.merge_from_branch(self.tree2.branch)
        self.tree.commit("second", rev_id=b"r2")

    def get_in_history(self, revision_spec):
        return spec_in_history(revision_spec, self.tree.branch)

    def assertInHistoryIs(self, exp_revno, exp_revision_id, revision_spec):
        rev_info = self.get_in_history(revision_spec)
        self.assertEqual(
            exp_revno,
            rev_info.revno,
            "Revision spec: {!r} returned wrong revno: {!r} != {!r}".format(
                revision_spec, exp_revno, rev_info.revno
            ),
        )
        self.assertEqual(
            exp_revision_id,
            rev_info.rev_id,
            "Revision spec: {!r} returned wrong revision id: {!r} != {!r}".format(
                revision_spec, exp_revision_id, rev_info.rev_id
            ),
        )

    def assertInvalid(self, revision_spec, extra="", invalid_as_revision_id=True):
        try:
            self.get_in_history(revision_spec)
        except InvalidRevisionSpec as e:
            self.assertEqual(revision_spec, e.spec)
            self.assertEqual(extra, e.extra)
        else:
            self.fail(
                "Expected InvalidRevisionSpec to be raised for {!r}.in_history".format(
                    revision_spec
                )
            )
        if invalid_as_revision_id:
            try:
                spec = RevisionSpec.from_string(revision_spec)
                spec.as_revision_id(self.tree.branch)
            except InvalidRevisionSpec as e:
                self.assertEqual(revision_spec, e.spec)
                self.assertEqual(extra, e.extra)
            else:
                self.fail(
                    "Expected InvalidRevisionSpec to be raised for"
                    " {!r}.as_revision_id".format(revision_spec)
                )

    def assertAsRevisionId(self, revision_id, revision_spec):
        """Calling as_revision_id() should return the specified id."""
        spec = RevisionSpec.from_string(revision_spec)
        self.assertEqual(revision_id, spec.as_revision_id(self.tree.branch))

    def get_as_tree(self, revision_spec, tree=None):
        if tree is None:
            tree = self.tree
        spec = RevisionSpec.from_string(revision_spec)
        return spec.as_tree(tree.branch)


class RevisionSpecMatchOnTrap(RevisionSpec):
    def _match_on(self, branch, revs):
        self.last_call = (branch, revs)
        return super()._match_on(branch, revs)


class TestRevisionSpecBase(TestRevisionSpec):
    def test_wants_no_revision_history(self):
        # If wants_revision_history = False, then _match_on should get None for
        # the branch revision history
        spec = RevisionSpecMatchOnTrap("foo", _internal=True)
        spec.in_history(self.tree.branch)

        self.assertEqual((self.tree.branch, None), spec.last_call)


class TestOddRevisionSpec(TestRevisionSpec):
    """Test things that aren't normally thought of as revision specs."""

    def test_none(self):
        self.assertInHistoryIs(None, None, None)

    def test_object(self):
        self.assertRaises(TypeError, RevisionSpec.from_string, object())


class RevisionSpec_bork(RevisionSpec):
    prefix = "irrelevant:"

    def _match_on(self, branch, revs):
        if self.spec == "bork":
            return RevisionInfo.from_revision_id(branch, b"r1")
        else:
            raise InvalidRevisionSpec(self.spec, branch)


class TestRevisionSpec_dwim(TestRevisionSpec):
    # Don't need to test revno's explicitly since TRS_revno already
    # covers that well for us
    def test_dwim_spec_revno(self):
        self.assertInHistoryIs(2, b"r2", "2")
        self.assertAsRevisionId(b"alt_r2", "1.1.1")

    def test_dwim_spec_revid(self):
        self.assertInHistoryIs(2, b"r2", "r2")

    def test_dwim_spec_tag(self):
        self.tree.branch.tags.set_tag("footag", b"r1")
        self.assertAsRevisionId(b"r1", "footag")
        self.tree.branch.tags.delete_tag("footag")
        self.assertRaises(InvalidRevisionSpec, self.get_in_history, "footag")

    def test_dwim_spec_tag_that_looks_like_revno(self):
        # Test that we slip past revno with things that look like revnos,
        # but aren't.  Tags are convenient for testing this since we can
        # make them look however we want.
        self.tree.branch.tags.set_tag("3", b"r2")
        self.assertAsRevisionId(b"r2", "3")
        self.build_tree(["tree/b"])
        self.tree.add(["b"])
        self.tree.commit("b", rev_id=b"r3")
        self.assertAsRevisionId(b"r3", "3")

    def test_dwim_spec_date(self):
        self.assertAsRevisionId(b"r1", "today")

    def test_dwim_spec_branch(self):
        self.assertInHistoryIs(None, b"alt_r2", "tree2")

    def test_dwim_spec_nonexistent(self):
        self.assertInvalid("somethingrandom", invalid_as_revision_id=False)
        self.assertInvalid("-1.1", invalid_as_revision_id=False)
        self.assertInvalid(".1", invalid_as_revision_id=False)
        self.assertInvalid("1..1", invalid_as_revision_id=False)
        self.assertInvalid("1.2..1", invalid_as_revision_id=False)
        self.assertInvalid("1.", invalid_as_revision_id=False)

    def test_append_dwim_revspec(self):
        original_dwim_revspecs = list(RevisionSpec_dwim._possible_revspecs)

        def reset_dwim_revspecs():
            RevisionSpec_dwim._possible_revspecs = original_dwim_revspecs

        self.addCleanup(reset_dwim_revspecs)
        RevisionSpec_dwim.append_possible_revspec(RevisionSpec_bork)
        self.assertAsRevisionId(b"r1", "bork")

    def test_append_lazy_dwim_revspec(self):
        original_dwim_revspecs = list(RevisionSpec_dwim._possible_revspecs)

        def reset_dwim_revspecs():
            RevisionSpec_dwim._possible_revspecs = original_dwim_revspecs

        self.addCleanup(reset_dwim_revspecs)
        RevisionSpec_dwim.append_possible_lazy_revspec(
            "breezy.tests.test_revisionspec", "RevisionSpec_bork"
        )
        self.assertAsRevisionId(b"r1", "bork")


class TestRevisionSpec_revno(TestRevisionSpec):
    def test_positive_int(self):
        self.assertInHistoryIs(0, b"null:", "0")
        self.assertInHistoryIs(1, b"r1", "1")
        self.assertInHistoryIs(2, b"r2", "2")
        self.assertInvalid("3")

    def test_dotted_decimal(self):
        self.assertInHistoryIs(None, b"alt_r2", "1.1.1")
        self.assertInvalid("1.1.123")

    def test_negative_int(self):
        self.assertInHistoryIs(2, b"r2", "-1")
        self.assertInHistoryIs(1, b"r1", "-2")

        self.assertInHistoryIs(1, b"r1", "-3")
        self.assertInHistoryIs(1, b"r1", "-4")
        self.assertInHistoryIs(1, b"r1", "-100")

    def test_positive(self):
        self.assertInHistoryIs(0, b"null:", "revno:0")
        self.assertInHistoryIs(1, b"r1", "revno:1")
        self.assertInHistoryIs(2, b"r2", "revno:2")

        self.assertInvalid("revno:3")

    def test_negative(self):
        self.assertInHistoryIs(2, b"r2", "revno:-1")
        self.assertInHistoryIs(1, b"r1", "revno:-2")

        self.assertInHistoryIs(1, b"r1", "revno:-3")
        self.assertInHistoryIs(1, b"r1", "revno:-4")

    def test_invalid_number(self):
        # Get the right exception text
        try:
            int("X")
        except ValueError as e:
            self.assertInvalid("revno:X", extra="\n" + str(e))
        else:
            self.fail()

    def test_missing_number_and_branch(self):
        self.assertInvalid(
            "revno::", extra="\ncannot have an empty revno and no branch"
        )

    def test_invalid_number_with_branch(self):
        try:
            int("X")
        except ValueError as e:
            self.assertInvalid("revno:X:tree2", extra="\n" + str(e))
        else:
            self.fail()

    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        spec = RevisionSpec.from_string("revno:2:tree2/a")
        self.assertRaises(errors.NotBranchError, spec.in_history, self.tree.branch)

    def test_with_branch(self):
        # Passing a URL overrides the supplied branch path
        revinfo = self.get_in_history("revno:2:tree2")
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b"alt_r2", revinfo.rev_id)

    def test_int_with_branch(self):
        revinfo = self.get_in_history("2:tree2")
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b"alt_r2", revinfo.rev_id)

    def test_with_url(self):
        url = self.get_url() + "/tree2"
        revinfo = self.get_in_history("revno:2:{}".format(url))
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b"alt_r2", revinfo.rev_id)

    def test_negative_with_url(self):
        url = self.get_url() + "/tree2"
        revinfo = self.get_in_history("revno:-1:{}".format(url))
        self.assertNotEqual(self.tree.branch.base, revinfo.branch.base)
        self.assertEqual(self.tree2.branch.base, revinfo.branch.base)
        self.assertEqual(2, revinfo.revno)
        self.assertEqual(b"alt_r2", revinfo.rev_id)

    def test_different_history_lengths(self):
        # Make sure we use the revisions and offsets in the supplied branch
        # not the ones in the original branch.
        self.tree2.commit("three", rev_id=b"r3")
        self.assertInHistoryIs(3, b"r3", "revno:3:tree2")
        self.assertInHistoryIs(3, b"r3", "revno:-1:tree2")

    def test_invalid_branch(self):
        self.assertRaises(errors.NotBranchError, self.get_in_history, "revno:-1:tree3")

    def test_invalid_revno_in_branch(self):
        self.tree.commit("three", rev_id=b"r3")
        self.assertInvalid("revno:3:tree2")

    def test_revno_n_path(self):
        """Old revno:N:path tests."""
        wta = self.make_branch_and_tree("a")
        ba = wta.branch

        wta.commit("Commit one", rev_id=b"a@r-0-1")
        wta.commit("Commit two", rev_id=b"a@r-0-2")
        wta.commit("Commit three", rev_id=b"a@r-0-3")

        wtb = self.make_branch_and_tree("b")
        bb = wtb.branch

        wtb.commit("Commit one", rev_id=b"b@r-0-1")
        wtb.commit("Commit two", rev_id=b"b@r-0-2")
        wtb.commit("Commit three", rev_id=b"b@r-0-3")

        self.assertEqual((1, b"a@r-0-1"), spec_in_history("revno:1:a/", ba))
        # The argument of in_history should be ignored since it is
        # redundant with the path in the spec.
        self.assertEqual((1, b"a@r-0-1"), spec_in_history("revno:1:a/", None))
        self.assertEqual((1, b"a@r-0-1"), spec_in_history("revno:1:a/", bb))
        self.assertEqual((2, b"b@r-0-2"), spec_in_history("revno:2:b/", None))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"null:", "0")
        self.assertAsRevisionId(b"r1", "1")
        self.assertAsRevisionId(b"r2", "2")
        self.assertAsRevisionId(b"r1", "-2")
        self.assertAsRevisionId(b"r2", "-1")
        self.assertAsRevisionId(b"alt_r2", "1.1.1")

    def test_as_tree(self):
        tree = self.get_as_tree("0")
        self.assertEqual(_mod_revision.NULL_REVISION, tree.get_revision_id())
        tree = self.get_as_tree("1")
        self.assertEqual(b"r1", tree.get_revision_id())
        tree = self.get_as_tree("2")
        self.assertEqual(b"r2", tree.get_revision_id())
        tree = self.get_as_tree("-2")
        self.assertEqual(b"r1", tree.get_revision_id())
        tree = self.get_as_tree("-1")
        self.assertEqual(b"r2", tree.get_revision_id())
        tree = self.get_as_tree("1.1.1")
        self.assertEqual(b"alt_r2", tree.get_revision_id())


class TestRevisionSpec_revid(TestRevisionSpec):
    def test_in_history(self):
        # We should be able to access revisions that are directly
        # in the history.
        self.assertInHistoryIs(1, b"r1", "revid:r1")
        self.assertInHistoryIs(2, b"r2", "revid:r2")

    def test_missing(self):
        self.assertInvalid("revid:r3", invalid_as_revision_id=False)

    def test_merged(self):
        """We can reach revisions in the ancestry."""
        self.assertInHistoryIs(None, b"alt_r2", "revid:alt_r2")

    def test_not_here(self):
        self.tree2.commit("alt third", rev_id=b"alt_r3")
        # It exists in tree2, but not in tree
        self.assertInvalid("revid:alt_r3", invalid_as_revision_id=False)

    def test_in_repository(self):
        """We can get any revision id in the repository."""
        # XXX: This may change in the future, but for now, it is true
        self.tree2.commit("alt third", rev_id=b"alt_r3")
        self.tree.branch.fetch(self.tree2.branch, b"alt_r3")
        self.assertInHistoryIs(None, b"alt_r3", "revid:alt_r3")

    def test_unicode(self):
        """We correctly convert a unicode ui string to an encoded revid."""
        revision_id = "\N{SNOWMAN}".encode()
        self.tree.commit("unicode", rev_id=revision_id)
        self.assertInHistoryIs(3, revision_id, "revid:\N{SNOWMAN}")
        self.assertInHistoryIs(3, revision_id, "revid:" + revision_id.decode("utf-8"))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"r1", "revid:r1")
        self.assertAsRevisionId(b"r2", "revid:r2")
        self.assertAsRevisionId(b"alt_r2", "revid:alt_r2")


class TestRevisionSpec_last(TestRevisionSpec):
    def test_positive(self):
        self.assertInHistoryIs(2, b"r2", "last:1")
        self.assertInHistoryIs(1, b"r1", "last:2")
        self.assertInHistoryIs(0, b"null:", "last:3")

    def test_empty(self):
        self.assertInHistoryIs(2, b"r2", "last:")

    def test_negative(self):
        self.assertInvalid("last:-1", extra="\nyou must supply a positive value")

    def test_missing(self):
        self.assertInvalid("last:4")

    def test_no_history(self):
        tree = self.make_branch_and_tree("tree3")

        self.assertRaises(errors.NoCommits, spec_in_history, "last:", tree.branch)

    def test_not_a_number(self):
        last_e = None
        try:
            int("Y")
        except ValueError as e:
            last_e = e
        self.assertInvalid("last:Y", extra="\n" + str(last_e))

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"r2", "last:1")
        self.assertAsRevisionId(b"r1", "last:2")


class TestRevisionSpec_before(TestRevisionSpec):
    def test_int(self):
        self.assertInHistoryIs(1, b"r1", "before:2")
        self.assertInHistoryIs(1, b"r1", "before:-1")

    def test_before_one(self):
        self.assertInHistoryIs(0, b"null:", "before:1")

    def test_before_none(self):
        self.assertInvalid("before:0", extra="\ncannot go before the null: revision")

    def test_revid(self):
        self.assertInHistoryIs(1, b"r1", "before:revid:r2")

    def test_last(self):
        self.assertInHistoryIs(1, b"r1", "before:last:1")

    def test_alt_revid(self):
        # This will grab the left-most ancestor for alternate histories
        self.assertInHistoryIs(1, b"r1", "before:revid:alt_r2")

    def test_alt_no_parents(self):
        new_tree = self.make_branch_and_tree("new_tree")
        new_tree.commit("first", rev_id=b"new_r1")
        self.tree.branch.fetch(new_tree.branch, b"new_r1")
        self.assertInHistoryIs(0, b"null:", "before:revid:new_r1")

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"r1", "before:revid:r2")
        self.assertAsRevisionId(b"r1", "before:2")
        self.assertAsRevisionId(b"r1", "before:1.1.1")
        self.assertAsRevisionId(b"r1", "before:revid:alt_r2")


class TestRevisionSpec_tag(TestRevisionSpec):
    def make_branch_and_tree(self, relpath):
        # override format as the default one may not support tags
        return TestRevisionSpec.make_branch_and_tree(
            self, relpath, format="dirstate-tags"
        )

    def test_from_string_tag(self):
        spec = RevisionSpec.from_string("tag:bzr-0.14")
        self.assertIsInstance(spec, RevisionSpec_tag)
        self.assertEqual(spec.spec, "bzr-0.14")

    def test_lookup_tag(self):
        self.tree.branch.tags.set_tag("bzr-0.14", b"r1")
        self.assertInHistoryIs(1, b"r1", "tag:bzr-0.14")
        self.tree.branch.tags.set_tag("null_rev", b"null:")
        self.assertInHistoryIs(0, b"null:", "tag:null_rev")

    def test_failed_lookup(self):
        # tags that don't exist give a specific message: arguably we should
        # just give InvalidRevisionSpec but I think this is more helpful
        self.assertRaises(errors.NoSuchTag, self.get_in_history, "tag:some-random-tag")

    def test_as_revision_id(self):
        self.tree.branch.tags.set_tag("my-tag", b"r2")
        self.tree.branch.tags.set_tag("null_rev", b"null:")
        self.assertAsRevisionId(b"r2", "tag:my-tag")
        self.assertAsRevisionId(b"null:", "tag:null_rev")
        self.assertAsRevisionId(b"r1", "before:tag:my-tag")


class TestRevisionSpec_date(TestRevisionSpec):
    def setUp(self):
        super(TestRevisionSpec, self).setUp()

        new_tree = self.make_branch_and_tree("new_tree")
        self.revid1 = new_tree.commit(
            "Commit one", timestamp=time.time() - 60 * 60 * 24
        )
        self.revid2 = new_tree.commit("Commit two")
        self.revid3 = new_tree.commit("Commit three")

        self.tree = new_tree

    def test_tomorrow(self):
        self.assertInvalid("date:tomorrow")

    def test_today(self):
        self.assertInHistoryIs(2, self.revid2, "date:today")
        self.assertInHistoryIs(1, self.revid1, "before:date:today")

    def test_yesterday(self):
        self.assertInHistoryIs(1, self.revid1, "date:yesterday")

    def test_invalid(self):
        self.assertInvalid("date:foobar", extra="\ninvalid date")
        # You must have '-' between year/month/day
        self.assertInvalid("date:20040404", extra="\ninvalid date")
        # Need 2 digits for each date piece
        self.assertInvalid("date:2004-4-4", extra="\ninvalid date")

    def test_day(self):
        now = datetime.datetime.now()
        self.assertInHistoryIs(
            2, self.revid2, "date:%04d-%02d-%02d" % (now.year, now.month, now.day)
        )

    def test_as_revision_id(self):
        self.assertAsRevisionId(self.revid2, "date:today")


class TestRevisionSpec_date_no_revno(TestRevisionSpec_date):
    # some formats don't implement .revno(), so it triggers a different codepath

    def get_in_history(self, revision_spec):
        old_revno = self.overrideAttr(self.tree.branch, "revno", lambda: None)
        try:
            return spec_in_history(revision_spec, self.tree.branch)
        finally:
            self.tree.branch.revno = old_revno

    def test_today(self):
        self.assertInHistoryIs(2, self.revid2, "date:today")
        # Drop before: since it messes with our monkeypatching of Branch.revno.


class TestRevisionSpec_ancestor(TestRevisionSpec):
    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        self.assertRaises(
            errors.NotBranchError, self.get_in_history, "ancestor:tree2/a"
        )

    def test_simple(self):
        # Common ancestor of trees is 'alt_r2'
        self.assertInHistoryIs(None, b"alt_r2", "ancestor:tree2")

        # Going the other way, we get a valid revno
        tmp = self.tree
        self.tree = self.tree2
        self.tree2 = tmp
        self.assertInHistoryIs(2, b"alt_r2", "ancestor:tree")

    def test_self(self):
        self.assertInHistoryIs(2, b"r2", "ancestor:tree")

    def test_unrelated(self):
        new_tree = self.make_branch_and_tree("new_tree")

        new_tree.commit("Commit one", rev_id=b"new_r1")
        new_tree.commit("Commit two", rev_id=b"new_r2")
        new_tree.commit("Commit three", rev_id=b"new_r3")

        # With no common ancestor, we should raise another user error
        self.assertRaises(
            errors.NoCommonAncestor, self.get_in_history, "ancestor:new_tree"
        )

    def test_no_commits(self):
        new_tree = self.make_branch_and_tree("new_tree")
        self.assertRaises(
            errors.NoCommits, spec_in_history, "ancestor:new_tree", self.tree.branch
        )

        self.assertRaises(
            errors.NoCommits, spec_in_history, "ancestor:tree", new_tree.branch
        )

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"alt_r2", "ancestor:tree2")

    def test_default(self):
        # We don't have a parent to default to
        self.assertRaises(errors.NotBranchError, self.get_in_history, "ancestor:")

        # Create a branch with a parent to default to
        tree3 = self.tree.controldir.sprout("tree3").open_workingtree()
        tree3.commit("foo", rev_id=b"r3")
        self.tree = tree3
        self.assertInHistoryIs(2, b"r2", "ancestor:")


class TestRevisionSpec_branch(TestRevisionSpec):
    def test_non_exact_branch(self):
        # It seems better to require an exact path to the branch
        # Branch.open() rather than using Branch.open_containing()
        self.assertRaises(errors.NotBranchError, self.get_in_history, "branch:tree2/a")

    def test_simple(self):
        self.assertInHistoryIs(None, b"alt_r2", "branch:tree2")

    def test_self(self):
        self.assertInHistoryIs(2, b"r2", "branch:tree")

    def test_unrelated(self):
        new_tree = self.make_branch_and_tree("new_tree")

        new_tree.commit("Commit one", rev_id=b"new_r1")
        new_tree.commit("Commit two", rev_id=b"new_r2")
        new_tree.commit("Commit three", rev_id=b"new_r3")

        self.assertInHistoryIs(None, b"new_r3", "branch:new_tree")

        # XXX: Right now, we use fetch() to make sure the remote revisions
        # have been pulled into the local branch. We may change that
        # behavior in the future.
        self.assertTrue(self.tree.branch.repository.has_revision(b"new_r3"))

    def test_no_commits(self):
        self.make_branch_and_tree("new_tree")
        self.assertRaises(errors.NoCommits, self.get_in_history, "branch:new_tree")
        self.assertRaises(errors.NoCommits, self.get_as_tree, "branch:new_tree")

    def test_as_revision_id(self):
        self.assertAsRevisionId(b"alt_r2", "branch:tree2")

    def test_as_tree(self):
        tree = self.get_as_tree("branch:tree", self.tree2)
        self.assertEqual(b"r2", tree.get_revision_id())
        self.assertFalse(self.tree2.branch.repository.has_revision(b"r2"))


class TestRevisionSpec_submit(TestRevisionSpec):
    def test_submit_branch(self):
        # Common ancestor of trees is 'alt_r2'
        self.assertRaises(errors.NoSubmitBranch, self.get_in_history, "submit:")
        self.tree.branch.set_parent("../tree2")
        self.assertInHistoryIs(None, b"alt_r2", "submit:")
        self.tree.branch.set_parent("bogus")
        self.assertRaises(errors.NotBranchError, self.get_in_history, "submit:")
        # submit branch overrides parent branch
        self.tree.branch.set_submit_branch("tree2")
        self.assertInHistoryIs(None, b"alt_r2", "submit:")

    def test_as_revision_id(self):
        self.tree.branch.set_submit_branch("tree2")
        self.assertAsRevisionId(b"alt_r2", "branch:tree2")


class TestRevisionSpec_mainline(TestRevisionSpec):
    def test_as_revision_id(self):
        self.assertAsRevisionId(b"r1", "mainline:1")
        self.assertAsRevisionId(b"r2", "mainline:1.1.1")
        self.assertAsRevisionId(b"r2", "mainline:revid:alt_r2")
        spec = RevisionSpec.from_string("mainline:revid:alt_r22")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            "Requested revision: 'mainline:revid:alt_r22' does not exist in branch: ",
        )

    def test_in_history(self):
        self.assertInHistoryIs(2, b"r2", "mainline:revid:alt_r2")


class TestRevisionSpec_annotate(TestRevisionSpec):
    def setUp(self):
        super().setUp()
        self.tree = self.make_branch_and_tree("annotate-tree")
        self.build_tree_contents([("annotate-tree/file1", b"1\n")])
        self.tree.add("file1")
        self.tree.commit("r1", rev_id=b"r1")
        self.build_tree_contents([("annotate-tree/file1", b"2\n1\n")])
        self.tree.commit("r2", rev_id=b"r2")
        self.build_tree_contents([("annotate-tree/file1", b"2\n1\n3\n")])

    def test_as_revision_id_r1(self):
        self.assertAsRevisionId(b"r1", "annotate:annotate-tree/file1:2")

    def test_as_revision_id_r2(self):
        self.assertAsRevisionId(b"r2", "annotate:annotate-tree/file1:1")

    def test_as_revision_id_uncommitted(self):
        spec = RevisionSpec.from_string("annotate:annotate-tree/file1:3")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            r"Requested revision: \'annotate:annotate-tree/file1:3\' does not"
            " exist in branch: .*\nLine 3 has not been committed.",
        )

    def test_non_existent_line(self):
        spec = RevisionSpec.from_string("annotate:annotate-tree/file1:4")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            r"Requested revision: \'annotate:annotate-tree/file1:4\' does not"
            " exist in branch: .*\nNo such line: 4",
        )

    def test_invalid_line(self):
        spec = RevisionSpec.from_string("annotate:annotate-tree/file1:q")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            r"Requested revision: \'annotate:annotate-tree/file1:q\' does not"
            " exist in branch: .*\nNo such line: q",
        )

    def test_no_such_file(self):
        spec = RevisionSpec.from_string("annotate:annotate-tree/file2:1")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            r"Requested revision: \'annotate:annotate-tree/file2:1\' does not"
            " exist in branch: .*\nFile 'file2' is not versioned",
        )

    def test_no_such_file_with_colon(self):
        spec = RevisionSpec.from_string("annotate:annotate-tree/fi:le2:1")
        e = self.assertRaises(
            InvalidRevisionSpec, spec.as_revision_id, self.tree.branch
        )
        self.assertContainsRe(
            str(e),
            r"Requested revision: \'annotate:annotate-tree/fi:le2:1\' does not"
            " exist in branch: .*\nFile 'fi:le2' is not versioned",
        )



# Copyright (C) 2018 Breezy Developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Support for GitHub."""

import json
import os
from datetime import datetime
from typing import Any, Optional

from ... import bedding, controldir, errors, urlutils
from ... import branch as _mod_branch
from ...config import AuthenticationConfig
from ...errors import PermissionDenied, UnexpectedHttpStatus
from ...forge import (
    Forge,
    ForgeLoginRequired,
    MergeProposal,
    MergeProposalBuilder,
    MergeProposalExists,
    NoSuchProject,
    PrerequisiteBranchUnsupported,
    ReopenFailed,
    UnsupportedForge,
    determine_title,
)
from ...git.urls import git_url_to_bzr_url
from ...i18n import gettext
from ...trace import note
from ...transport import get_transport

GITHUB_HOST = "github.com"
WEB_GITHUB_URL = "https://github.com"
API_GITHUB_URL = "https://api.github.com"
DEFAULT_PER_PAGE = 100

SCHEME_FIELD_MAP = {
    "ssh": "ssh_url",
    "git+ssh": "ssh_url",
    "http": "clone_url",
    "https": "clone_url",
    "git": "git_url",
}
DEFAULT_PREFERRED_SCHEMES = ["ssh", "http"]


def parse_timestring(ts):
    return datetime.strptime(ts, "%Y-%m-%dT%H:%M:%SZ")


def store_github_token(token):
    auth_config = AuthenticationConfig()
    auth_config._set_option("Github", "scheme", "https")
    auth_config._set_option("Github", "url", API_GITHUB_URL)
    auth_config._set_option("Github", "private_token", token)


def retrieve_github_token():
    auth_config = AuthenticationConfig()
    section = auth_config._get_config().get("Github")
    if section and section.get("private_token"):
        return section.get("private_token")

    # Backwards compatibility
    path = os.path.join(bedding.config_dir(), "github.conf")
    if not os.path.exists(path):
        return None
    with open(path) as f:
        return f.read().strip()


class ValidationFailed(errors.BzrError):
    _fmt = "GitHub validation failed: %(error)s"

    def __init__(self, error):
        errors.BzrError.__init__(self)
        self.error = error


class NotGitHubUrl(errors.BzrError):
    _fmt = "Not a GitHub URL: %(url)s"

    def __init__(self, url):
        errors.BzrError.__init__(self)
        self.url = url


class GitHubLoginRequired(ForgeLoginRequired):
    _fmt = "Action requires GitHub login."


class GitHubMergeProposal(MergeProposal):
    supports_auto_merge = True

    def __init__(self, gh, pr):
        self._gh = gh
        self._pr = pr

    def __repr__(self):
        return "<{} at {!r}>".format(type(self).__name__, self.url)

    name = "GitHub"

    def get_web_url(self):
        return self._pr["html_url"]

    @property
    def url(self):
        return self._pr["html_url"]

    def _branch_from_part(self, part, preferred_schemes=None):
        if part["repo"] is None:
            return None
        if preferred_schemes is None:
            preferred_schemes = DEFAULT_PREFERRED_SCHEMES
        for scheme in preferred_schemes:
            if scheme in SCHEME_FIELD_MAP:
                return github_url_to_bzr_url(
                    part["repo"][SCHEME_FIELD_MAP[scheme]], part["ref"]
                )
        raise AssertionError

    def get_source_branch_url(self, *, preferred_schemes=None):
        return self._branch_from_part(
            self._pr["head"], preferred_schemes=preferred_schemes
        )

    def get_source_revision(self):
        """Return the latest revision for the source branch."""
        from breezy.git.mapping import default_mapping

        return default_mapping.revision_id_foreign_to_bzr(
            self._pr["head"]["sha"].encode("ascii")
        )

    def get_target_branch_url(self, *, preferred_schemes=None):
        return self._branch_from_part(
            self._pr["base"], preferred_schemes=preferred_schemes
        )

    def set_target_branch_name(self, name):
        self._patch(base=name)

    def get_source_project(self):
        if self._pr["head"]["repo"] is None:
            return None
        return self._pr["head"]["repo"]["full_name"]

    def get_target_project(self):
        if self._pr["base"]["repo"] is None:
            return None
        return self._pr["base"]["repo"]["full_name"]

    def get_description(self):
        return self._pr["body"]

    def get_commit_message(self):
        return None

    def get_title(self):
        return self._pr.get("title")

    def set_title(self, title):
        self._patch(title=title)

    def set_commit_message(self, message):
        raise errors.UnsupportedOperation(self.set_commit_message, self)

    def _patch(self, **data):
        response = self._gh._api_request(
            "PATCH", self._pr["url"], body=json.dumps(data).encode("utf-8")
        )
        if response.status == 422:
            raise ValidationFailed(json.loads(response.text))
        if response.status != 200:
            raise UnexpectedHttpStatus(
                self._pr["url"], response.status, headers=response.getheaders()
            )
        self._pr = json.loads(response.text)

    def set_description(self, description):
        self._patch(body=description, title=determine_title(description))

    def is_merged(self):
        return bool(self._pr.get("merged_at"))

    def is_closed(self):
        return self._pr["state"] == "closed" and not bool(self._pr.get("merged_at"))

    def reopen(self):
        try:
            self._patch(state="open")
        except ValidationFailed as e:
            raise ReopenFailed(e.error["errors"][0]["message"])

    def close(self):
        self._patch(state="closed")

    def can_be_merged(self):
        return self._pr["mergeable"]

    def merge(self, commit_message=None, auto=False):
        if auto:
            graphql_query = """
mutation ($pullRequestId: ID!) {
  enablePullRequestAutoMerge(input: {
    pullRequestId: $pullRequestId,
    mergeMethod: MERGE
  }) {
    pullRequest {
      autoMergeRequest {
        enabledAt
        enabledBy {
          login
        }
      }
    }
  }
}
"""
            try:
                self._gh._graphql_request(
                    graphql_query, pullRequestId=self._pr["node_id"]
                )
            except GraphqlErrors as e:
                mutter("graphql errors: %r", e.errors)
                first_error = e.errors[0]
                if (
                    first_error["type"] == "UNPROCESSABLE"
                    and first_error["path"] == "enablePullRequestAutoMerge"
                ):
                    # TODO(jelmer): better exception type
                    raise Exception(first_error["message"])
                raise Exception(first_error["message"])
        else:
            # https://developer.github.com/v3/pulls/#merge-a-pull-request-merge-button
            data = {}
            if commit_message:
                data["commit_message"] = commit_message
            response = self._gh._api_request(
                "PUT", self._pr["url"] + "/merge", body=json.dumps(data).encode("utf-8")
            )
            if response.status == 422:
                raise ValidationFailed(json.loads(response.text))
            if response.status != 200:
                raise UnexpectedHttpStatus(
                    self._pr["url"], response.status, headers=response.getheaders()
                )

    def get_merged_by(self):
        merged_by = self._pr.get("merged_by")
        if merged_by is None:
            return None
        return merged_by["login"]

    def get_merged_at(self):
        merged_at = self._pr.get("merged_at")
        if merged_at is None:
            return None
        return parse_timestring(merged_at)

    def post_comment(self, body):
        data = {"body": body}
        response = self._gh._api_request(
            "POST", self._pr["comments_url"], body=json.dumps(data).encode("utf-8")
        )
        if response.status == 422:
            raise ValidationFailed(json.loads(response.text))
        if response.status != 201:
            raise UnexpectedHttpStatus(
                self._pr["comments_url"], response.status, headers=response.getheaders()
            )
        json.loads(response.text)


def parse_github_url(url):
    (scheme, user, password, host, port, path) = urlutils.parse_url(url)
    if host != GITHUB_HOST:
        raise NotGitHubUrl(url)
    (owner, repo_name) = path.strip("/").split("/")
    if repo_name.endswith(".git"):
        repo_name = repo_name[:-4]
    return owner, repo_name


def parse_github_branch_url(branch):
    url = urlutils.strip_segment_parameters(branch.user_url)
    owner, repo_name = parse_github_url(url)
    return owner, repo_name, branch.name


def parse_github_pr_url(url):
    (scheme, user, password, host, port, path) = urlutils.parse_url(url)
    if host != GITHUB_HOST:
        raise NotGitHubUrl(url)
    try:
        (owner, repo_name, pull, pr_id) = path.strip("/").split("/")
    except IndexError as e:
        raise ValueError("Not a PR URL") from e

    if pull != "pull":
        raise ValueError("Not a PR URL")

    return (owner, repo_name, pr_id)


def github_url_to_bzr_url(url, branch_name):
    return git_url_to_bzr_url(url, branch_name)


def strip_optional(url):
    return url.split("{")[0]


class _LazyDict(dict):
    def __init__(self, base, load_fn):
        self._load_fn = load_fn
        super().update(base)

    def _load_full(self):
        super().update(self._load_fn())
        self._load_fn = None

    def __getitem__(self, key):
        if self._load_fn is not None:
            try:
                return super().__getitem__(key)
            except KeyError:
                self._load_full()
        return super().__getitem__(key)

    def items(self):
        self._load_full()
        return super().items()

    def keys(self):
        self._load_full()
        return super().keys()

    def values(self):
        self._load_full()
        return super().values()

    def __contains__(self, key):
        if super().__contains__(key):
            return True
        if self._load_fn is not None:
            self._load_full()
            return super().__contains__(key)
        return False

    def __delitem__(self, name):
        raise NotImplementedError

    def __setitem__(self, name, value):
        raise NotImplementedError

    def get(self, name, default=None):
        if self._load_fn is not None:
            try:
                return super().get(name, default)
            except KeyError:
                self._load_full()
        return super().get(name, default)

    def pop(self):
        raise NotImplementedError

    def popitem(self):
        raise NotImplementedError

    def clear(self):
        raise NotImplementedError


class GraphqlErrors(Exception):
    def __init__(self, errors):
        self.errors = errors


class GitHub(Forge):
    name = "github"

    supports_merge_proposal_labels = True
    supports_merge_proposal_commit_message = False
    supports_merge_proposal_title = True
    supports_allow_collaboration = True
    merge_proposal_description_format = "markdown"

    def __repr__(self):
        return "GitHub()"

    def _graphql_request(self, body, **kwargs):
        headers = {}
        if self._token:
            headers["Authorization"] = "token {}".format(self._token)
        url = urlutils.join(self.transport.base, "graphql")
        response = self.transport.request(
            "POST",
            url,
            headers=headers,
            body=json.dumps(
                {
                    "query": body,
                    "variables": kwargs,
                }
            ).encode("utf-8"),
        )
        if response.status != 200:
            raise UnexpectedHttpStatus(
                url, response.status, headers=response.getheaders()
            )
        data = json.loads(response.text)
        if data.get("errors"):
            raise GraphqlErrors(data.get("errors"))
        return data["data"]

    def _api_request(self, method, path, body=None):
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/vnd.github.v3+json",
        }
        if self._token:
            headers["Authorization"] = "token {}".format(self._token)
        try:
            response = self.transport.request(
                method,
                urlutils.join(self.transport.base, path),
                headers=headers,
                body=body,
                retries=3,
            )
        except UnexpectedHttpStatus as e:
            if e.code == 401:
                raise GitHubLoginRequired(self.base_url)
            else:
                raise
        if response.status == 401:
            raise GitHubLoginRequired(self.base_url)
        return response

    def _get_repo(self, owner, repo):
        path = "repos/{}/{}".format(owner, repo)
        response = self._api_request("GET", path)
        if response.status == 404:
            raise NoSuchProject(path)
        if response.status == 200:
            return json.loads(response.text)
        raise UnexpectedHttpStatus(path, response.status, headers=response.getheaders())

    def _get_repo_pulls(self, path, head=None, state=None):
        path = path + "?"
        params = {}
        if head is not None:
            params["head"] = head
        if state is not None:
            params["state"] = state
        path += ";".join(
            ["{}={}".format(k, urlutils.quote(v)) for k, v in params.items()]
        )
        response = self._api_request("GET", path)
        if response.status == 404:
            raise NoSuchProject(path)
        if response.status == 200:
            return json.loads(response.text)
        raise UnexpectedHttpStatus(path, response.status, headers=response.getheaders())

    def _create_pull(
        self,
        path,
        title,
        head,
        base,
        body=None,
        labels=None,
        assignee=None,
        draft=False,
        maintainer_can_modify=False,
    ):
        data = {
            "title": title,
            "head": head,
            "base": base,
            "draft": draft,
            "maintainer_can_modify": maintainer_can_modify,
        }
        if labels is not None:
            data["labels"] = labels
        if assignee is not None:
            data["assignee"] = assignee
        if body:
            data["body"] = body

        response = self._api_request(
            "POST", path, body=json.dumps(data).encode("utf-8")
        )
        if response.status == 403:
            raise PermissionDenied(path, response.text)
        if response.status == 422:
            raise ValidationFailed(json.loads(response.text))
        if response.status != 201:
            raise UnexpectedHttpStatus(
                path, response.status, headers=response.getheaders()
            )
        return json.loads(response.text)

    def _get_user_by_email(self, email):
        path = "search/users?q={}+in:email".format(email)
        response = self._api_request("GET", path)
        if response.status != 200:
            raise UnexpectedHttpStatus(
                path, response.status, headers=response.getheaders()
            )
        ret = json.loads(response.text)
        if ret["total_count"] == 0:
            raise KeyError("no user with email {}".format(email))
        elif ret["total_count"] > 1:
            raise ValueError("more than one result for email {}".format(email))
        return ret["items"][0]

    def _get_user(self, username=None):
        if username:
            path = "users/{}".format(username)
        else:
            path = "user"
        response = self._api_request("GET", path)
        if response.status != 200:
            raise UnexpectedHttpStatus(
                path, response.status, headers=response.getheaders()
            )
        return json.loads(response.text)

    def _get_organization(self, name):
        path = "orgs/{}".format(name)
        response = self._api_request("GET", path)
        if response.status != 200:
            raise UnexpectedHttpStatus(
                path, response.status, headers=response.getheaders()
            )
        return json.loads(response.text)

    def _list_paged(self, path, parameters=None, per_page=None):
        if parameters is None:
            parameters = {}
        else:
            parameters = dict(parameters.items())
        if per_page:
            parameters["per_page"] = str(per_page)
        page = 1
        while path:
            parameters["page"] = str(page)
            response = self._api_request(
                "GET",
                path
                + "?"
                + ";".join(
                    [
                        "{}={}".format(k, urlutils.quote(v))
                        for (k, v) in parameters.items()
                    ]
                ),
            )
            if response.status != 200:
                raise UnexpectedHttpStatus(
                    path, response.status, headers=response.getheaders()
                )
            data = json.loads(response.text)
            if not data:
                break
            yield data
            page += 1

    def _search_issues(self, query):
        path = "search/issues"
        for page in self._list_paged(path, {"q": query}, per_page=DEFAULT_PER_PAGE):
            if not page["items"]:
                break
            yield from page["items"]

    def _create_fork(self, path, owner=None):
        if owner and owner != self.current_user["login"]:
            path += "?organization={}".format(owner)
        response = self._api_request("POST", path)
        if response.status != 202:
            raise UnexpectedHttpStatus(
                path, response.status, headers=response.getheaders()
            )
        return json.loads(response.text)

    @property
    def base_url(self):
        return WEB_GITHUB_URL

    def __init__(self, transport):
        self._token = retrieve_github_token()
        if self._token is None:
            note("Accessing GitHub anonymously. To log in, run 'brz gh-login'.")
        self.transport = transport
        self._current_user = None

    @property
    def current_user(self):
        if self._current_user is None:
            self._current_user = self._get_user()
        return self._current_user

    def publish_derived(
        self,
        local_branch,
        base_branch,
        name,
        project=None,
        owner=None,
        revision_id=None,
        overwrite=False,
        allow_lossy=True,
        tag_selector=None,
    ):
        if tag_selector is None:

            def tag_selector(t):
                return False

        base_owner, base_project, base_branch_name = parse_github_branch_url(
            base_branch
        )
        base_repo = self._get_repo(base_owner, base_project)
        if owner is None:
            owner = self.current_user["login"]
        if project is None:
            project = base_repo["name"]
        try:
            remote_repo = self._get_repo(owner, project)
        except NoSuchProject:
            base_repo = self._get_repo(base_owner, base_project)
            remote_repo = self._create_fork(base_repo["forks_url"], owner)
            note(
                gettext("Forking new repository %s from %s")
                % (remote_repo["html_url"], base_repo["html_url"])
            )
        else:
            note(gettext("Reusing existing repository %s") % remote_repo["html_url"])
        remote_dir = controldir.ControlDir.open(
            git_url_to_bzr_url(remote_repo["ssh_url"])
        )
        try:
            push_result = remote_dir.push_branch(
                local_branch,
                revision_id=revision_id,
                overwrite=overwrite,
                name=name,
                tag_selector=tag_selector,
            )
        except errors.NoRoundtrippingSupport:
            if not allow_lossy:
                raise
            push_result = remote_dir.push_branch(
                local_branch,
                revision_id=revision_id,
                overwrite=overwrite,
                name=name,
                lossy=True,
                tag_selector=tag_selector,
            )
        return push_result.target_branch, github_url_to_bzr_url(
            remote_repo["clone_url"], name
        )

    def get_push_url(self, branch):
        owner, project, branch_name = parse_github_branch_url(branch)
        repo = self._get_repo(owner, project)
        return github_url_to_bzr_url(repo["ssh_url"], branch_name)

    def get_web_url(self, branch):
        owner, project, branch_name = parse_github_branch_url(branch)
        repo = self._get_repo(owner, project)
        if branch_name:
            # TODO(jelmer): Don't hardcode this
            return repo["html_url"] + "/tree/" + branch_name
        else:
            return repo["html_url"]

    def get_derived_branch(
        self, base_branch, name, project=None, owner=None, preferred_schemes=None
    ):
        base_owner, base_project, base_branch_name = parse_github_branch_url(
            base_branch
        )
        base_repo = self._get_repo(base_owner, base_project)
        if owner is None:
            owner = self.current_user["login"]
        if project is None:
            project = base_repo["name"]
        try:
            remote_repo = self._get_repo(owner, project)
        except NoSuchProject:
            raise errors.NotBranchError(
                "{}/{}/{}".format(WEB_GITHUB_URL, owner, project)
            )
        if preferred_schemes is None:
            preferred_schemes = DEFAULT_PREFERRED_SCHEMES
        for scheme in preferred_schemes:
            if scheme in SCHEME_FIELD_MAP:
                github_url = remote_repo[SCHEME_FIELD_MAP[scheme]]
                break
        else:
            raise AssertionError
        full_url = github_url_to_bzr_url(github_url, name)
        return _mod_branch.Branch.open(full_url)

    def get_proposer(self, source_branch, target_branch):
        return GitHubMergeProposalBuilder(self, source_branch, target_branch)

    def iter_proposals(self, source_branch, target_branch, status="open"):
        (source_owner, source_repo_name, source_branch_name) = parse_github_branch_url(
            source_branch
        )
        (target_owner, target_repo_name, target_branch_name) = parse_github_branch_url(
            target_branch
        )
        target_repo = self._get_repo(target_owner, target_repo_name)
        state = {"open": "open", "merged": "closed", "closed": "closed", "all": "all"}
        pulls = self._get_repo_pulls(
            strip_optional(target_repo["pulls_url"]),
            head=target_branch_name,
            state=state[status],
        )
        for pull in pulls:
            if (status == "closed" and pull["merged"]) or (
                status == "merged" and not pull["merged"]
            ):
                continue
            if pull["head"]["ref"] != source_branch_name:
                continue
            if pull["head"]["repo"] is None:
                # Repo has gone the way of the dodo
                continue
            if (
                pull["head"]["repo"]["owner"]["login"] != source_owner
                or pull["head"]["repo"]["name"] != source_repo_name
            ):
                continue
            yield GitHubMergeProposal(self, pull)

    def hosts(self, branch):
        try:
            parse_github_branch_url(branch)
        except NotGitHubUrl:
            return False
        else:
            return True

    @classmethod
    def probe_from_hostname(cls, hostname, possible_transports=None):
        if hostname == GITHUB_HOST:
            transport = get_transport(
                API_GITHUB_URL, possible_transports=possible_transports
            )
            return cls(transport)
        raise UnsupportedForge(hostname)

    @classmethod
    def probe_from_url(cls, url, possible_transports=None):
        try:
            parse_github_url(url)
        except NotGitHubUrl:
            raise UnsupportedForge(url)
        transport = get_transport(
            API_GITHUB_URL, possible_transports=possible_transports
        )
        return cls(transport)

    @classmethod
    def iter_instances(cls):
        yield cls(get_transport(API_GITHUB_URL))

    def iter_my_proposals(self, status="open", author=None):
        query = ["is:pr"]
        if status == "open":
            query.append("is:open")
        elif status == "closed":
            query.append("is:unmerged")
            # Also use "is:closed" otherwise unmerged open pull requests are
            # also included.
            query.append("is:closed")
        elif status == "merged":
            query.append("is:merged")
        if author is None:
            author = self.current_user["login"]
        query.append("author:{}".format(author))
        for issue in self._search_issues(query=" ".join(query)):

            def retrieve_full():
                response = self._api_request("GET", issue["pull_request"]["url"])
                if response.status != 200:
                    raise UnexpectedHttpStatus(
                        issue["pull_request"]["url"],
                        response.status,
                        headers=response.getheaders(),
                    )
                return json.loads(response.text)

            yield GitHubMergeProposal(
                self, _LazyDict(issue["pull_request"], retrieve_full)
            )

    def get_proposal_by_url(self, url):
        try:
            (owner, repo, pr_id) = parse_github_pr_url(url)
        except NotGitHubUrl as e:
            raise UnsupportedForge(url) from e
        api_url = "https://api.github.com/repos/{}/{}/pulls/{}".format(
            owner, repo, pr_id
        )
        response = self._api_request("GET", api_url)
        if response.status != 200:
            raise UnexpectedHttpStatus(
                api_url, response.status, headers=response.getheaders()
            )
        data = json.loads(response.text)
        return GitHubMergeProposal(self, data)

    def iter_my_forks(self, owner=None):
        if owner:
            path = "/users/{}/repos".format(owner)
        else:
            path = "/user/repos"
        for page in self._list_paged(path, per_page=DEFAULT_PER_PAGE):
            for project in page:
                if not project["fork"]:
                    continue
                yield project["full_name"]

    def delete_project(self, path):
        path = "repos/" + path
        response = self._api_request("DELETE", path)
        if response.status == 404:
            raise NoSuchProject(path)
        if response.status == 204:
            return
        if response.status == 200:
            return json.loads(response.text)
        raise UnexpectedHttpStatus(path, response.status, headers=response.getheaders())

    def create_project(
        self,
        path,
        *,
        homepage=None,
        private=False,
        has_issues=True,
        has_projects=False,
        has_wiki=False,
        summary=None,
    ):
        owner, name = path.split("/")
        path = "repos"
        data = {
            "name": "name",
            "description": summary,
            "homepage": homepage,
            "private": private,
            "has_issues": has_issues,
            "has_projects": has_projects,
            "has_wiki": has_wiki,
        }
        response = self._api_request(
            "POST", path, body=json.dumps(data).encode("utf-8")
        )
        if response.status != 201:
            return json.loads(response.text)
        raise UnexpectedHttpStatus(path, response.status, headers=response.getheaders())

    def get_current_user(self):
        if self._token is not None:
            return self.current_user["login"]
        return None

    def get_user_url(self, username):
        return urlutils.join(self.base_url, username)


class GitHubMergeProposalBuilder(MergeProposalBuilder):
    def __init__(self, gh, source_branch, target_branch):
        self.gh = gh
        self.source_branch = source_branch
        self.target_branch = target_branch
        (self.target_owner, self.target_repo_name, self.target_branch_name) = (
            parse_github_branch_url(self.target_branch)
        )
        (self.source_owner, self.source_repo_name, self.source_branch_name) = (
            parse_github_branch_url(self.source_branch)
        )

    def get_infotext(self):
        """Determine the initial comment for the merge proposal."""
        info = []
        info.append(
            "Merge {} into {}:{}\n".format(
                self.source_branch_name, self.target_owner, self.target_branch_name
            )
        )
        info.append("Source: {}\n".format(self.source_branch.user_url))
        info.append("Target: {}\n".format(self.target_branch.user_url))
        return "".join(info)

    def get_initial_body(self):
        """Get a body for the proposal for the user to modify.

        :return: a str or None.
        """
        return None

    def create_proposal(
        self,
        description,
        title=None,
        reviewers=None,
        labels=None,
        prerequisite_branch=None,
        commit_message=None,
        work_in_progress=False,
        allow_collaboration=False,
        delete_source_after_merge: Optional[bool] = None,
    ):
        """Perform the submission."""
        if prerequisite_branch is not None:
            raise PrerequisiteBranchUnsupported(self)
        # Note that commit_message is ignored, since github doesn't support it.
        # TODO(jelmer): Probe for right repo name
        if self.target_repo_name.endswith(".git"):
            self.target_repo_name = self.target_repo_name[:-4]
        if title is None:
            title = determine_title(description)
        target_repo = self.gh._get_repo(self.target_owner, self.target_repo_name)
        assignees: Optional[list[dict[str, Any]]] = []
        if reviewers:
            assignees = []
            for reviewer in reviewers:
                if "@" in reviewer:
                    user = self.gh._get_user_by_email(reviewer)
                else:
                    user = self.gh._get_user(reviewer)
                assignees.append(user["login"])
        else:
            assignees = None
        kwargs: dict[str, Any] = {}
        if delete_source_after_merge is not None:
            kwargs["delete_branch_on_merge"] = delete_source_after_merge
        try:
            pull_request = self.gh._create_pull(
                strip_optional(target_repo["pulls_url"]),
                title=title,
                body=description,
                head="{}:{}".format(self.source_owner, self.source_branch_name),
                base=self.target_branch_name,
                labels=labels,
                assignee=assignees,
                draft=work_in_progress,
                maintainer_can_modify=allow_collaboration,
                **kwargs,
            )
        except ValidationFailed:
            # TODO(jelmer): Check the actual error message rather than assuming
            # a merge proposal exists?
            raise MergeProposalExists(self.source_branch.user_url)
        return GitHubMergeProposal(self.gh, pull_request)



# Copyright (C) 2005-2008, 2010 by Canonical Ltd
#   Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

from unittest import TestLoader

from .... import config, tests
from ....bzr.bzrdir import BzrDir
from ....tests import TestCaseInTempDir
from ..emailer import EmailSender


def test_suite():
    return TestLoader().loadTestsFromName(__name__)


sample_config = (
    b"[DEFAULT]\n"
    b"post_commit_to=demo@example.com\n"
    b"post_commit_sender=Sample <foo@example.com>\n"
    b"revision_mail_headers=X-Cheese: to the rescue!\n"
)

unconfigured_config = b"[DEFAULT]\nemail=Robert <foo@example.com>\n"

sender_configured_config = b"[DEFAULT]\npost_commit_sender=Sample <foo@example.com>\n"

to_configured_config = b"[DEFAULT]\npost_commit_to=Sample <foo@example.com>\n"

multiple_to_configured_config = (
    b"[DEFAULT]\n"
    b"post_commit_sender=Sender <from@example.com>\n"
    b"post_commit_to=Sample <foo@example.com>, Other <baz@bar.com>\n"
)

customized_mail_config = (
    b"[DEFAULT]\n"
    b"post_commit_to=demo@example.com\n"
    b"post_commit_sender=Sample <foo@example.com>\n"
    b"post_commit_subject=[commit] $message\n"
    b"post_commit_body='''$committer has committed "
    b"revision 1 at $url.\n\n'''\n"
)

push_config = (
    b"[DEFAULT]\npost_commit_to=demo@example.com\npost_commit_push_pull=True\n"
)

with_url_config = (
    b"[DEFAULT]\n"
    b"post_commit_url=http://some.fake/url/\n"
    b"post_commit_to=demo@example.com\n"
    b"post_commit_sender=Sample <foo@example.com>\n"
)

# FIXME: this should not use a literal log, rather grab one from breezy.log
sample_log = (
    "------------------------------------------------------------\n"
    "revno: 1\n"
    "revision-id: %s\n"
    "committer: Sample <john@example.com>\n"
    "branch nick: work\n"
    "timestamp: Thu 1970-01-01 00:00:01 +0000\n"
    "message:\n"
    "  foo bar baz\n"
    "  fuzzy\n"
    "  wuzzy\n"
)


class TestGetTo(TestCaseInTempDir):
    def test_body(self):
        sender, revid = self.get_sender()
        self.assertEqual(
            "At {}\n\n{}".format(sender.url(), sample_log % revid.decode("utf-8")),
            sender.body(),
        )

    def test_custom_body(self):
        sender, revid = self.get_sender(customized_mail_config)
        self.assertEqual(
            "{} has committed revision 1 at {}.\n\n{}".format(
                sender.revision.committer,
                sender.url(),
                sample_log % revid.decode("utf-8"),
            ),
            sender.body(),
        )

    def test_command_line(self):
        sender, revid = self.get_sender()
        self.assertEqual(
            ["mail", "-s", sender.subject(), "-a", "From: " + sender.from_address()]
            + sender.to(),
            sender._command_line(),
        )

    def test_to(self):
        sender, revid = self.get_sender()
        self.assertEqual(["demo@example.com"], sender.to())

    def test_from(self):
        sender, revid = self.get_sender()
        self.assertEqual("Sample <foo@example.com>", sender.from_address())

    def test_from_default(self):
        sender, revid = self.get_sender(unconfigured_config)
        self.assertEqual("Robert <foo@example.com>", sender.from_address())

    def test_should_send(self):
        sender, revid = self.get_sender()
        self.assertEqual(True, sender.should_send())

    def test_should_not_send(self):
        sender, revid = self.get_sender(unconfigured_config)
        self.assertEqual(False, sender.should_send())

    def test_should_not_send_sender_configured(self):
        sender, revid = self.get_sender(sender_configured_config)
        self.assertEqual(False, sender.should_send())

    def test_should_not_send_to_configured(self):
        sender, revid = self.get_sender(to_configured_config)
        self.assertEqual(True, sender.should_send())

    def test_send_to_multiple(self):
        sender, revid = self.get_sender(multiple_to_configured_config)
        self.assertEqual(
            ["Sample <foo@example.com>", "Other <baz@bar.com>"], sender.to()
        )
        self.assertEqual(
            ["Sample <foo@example.com>", "Other <baz@bar.com>"],
            sender._command_line()[-2:],
        )

    def test_url_set(self):
        sender, revid = self.get_sender(with_url_config)
        self.assertEqual(sender.url(), "http://some.fake/url/")

    def test_public_url_set(self):
        config = b"[DEFAULT]\npublic_branch=http://the.publication/location/\n"
        sender, revid = self.get_sender(config)
        self.assertEqual(sender.url(), "http://the.publication/location/")

    def test_url_precedence(self):
        config = (
            b"[DEFAULT]\n"
            b"post_commit_url=http://some.fake/url/\n"
            b"public_branch=http://the.publication/location/\n"
        )
        sender, revid = self.get_sender(config)
        self.assertEqual(sender.url(), "http://some.fake/url/")

    def test_url_unset(self):
        sender, revid = self.get_sender()
        self.assertEqual(sender.url(), sender.branch.base)

    def test_subject(self):
        sender, revid = self.get_sender()
        self.assertEqual(
            "Rev 1: foo bar baz in {}".format(sender.branch.base), sender.subject()
        )

    def test_custom_subject(self):
        sender, revid = self.get_sender(customized_mail_config)
        self.assertEqual(
            "[commit] {}".format(sender.revision.get_summary()), sender.subject()
        )

    def test_diff_filename(self):
        sender, revid = self.get_sender()
        self.assertEqual("patch-1.diff", sender.diff_filename())

    def test_headers(self):
        sender, revid = self.get_sender()
        self.assertEqual({"X-Cheese": "to the rescue!"}, sender.extra_headers())

    def get_sender(self, text=sample_config):
        my_config = config.MemoryStack(text)
        self.branch = BzrDir.create_branch_convenience(".")
        tree = self.branch.controldir.open_workingtree()
        revid = tree.commit(
            "foo bar baz\nfuzzy\nwuzzy",
            allow_pointless=True,
            timestamp=1,
            timezone=0,
            committer="Sample <john@example.com>",
        )
        sender = EmailSender(self.branch, revid, my_config)
        # This is usually only done after the EmailSender has locked the branch
        # and repository during send(), however, for testing, we need to do it
        # earlier, since send() is not called.
        sender._setup_revision_and_revno()
        return sender, revid


class TestEmailerWithLocal(tests.TestCaseWithTransport):
    """Test that Emailer will use a local branch if supplied."""

    def test_local_has_revision(self):
        master_tree = self.make_branch_and_tree("master")
        self.build_tree(["master/a"])
        master_tree.add("a")
        master_tree.commit("a")

        child_tree = master_tree.controldir.sprout("child").open_workingtree()
        child_tree.branch.bind(master_tree.branch)

        self.build_tree(["child/b"])
        child_tree.add(["b"])
        revision_id = child_tree.commit("b")

        sender = EmailSender(
            master_tree.branch,
            revision_id,
            master_tree.branch.get_config(),
            local_branch=child_tree.branch,
        )

        # Make sure we are using the 'local_branch' repository, and not the
        # remote one.
        self.assertIs(child_tree.branch.repository, sender.repository)

    def test_local_missing_revision(self):
        master_tree = self.make_branch_and_tree("master")
        self.build_tree(["master/a"])
        master_tree.add("a")
        master_tree.commit("a")

        child_tree = master_tree.controldir.sprout("child").open_workingtree()
        child_tree.branch.bind(master_tree.branch)

        self.build_tree(["master/c"])
        master_tree.add(["c"])
        revision_id = master_tree.commit("c")

        self.assertFalse(child_tree.branch.repository.has_revision(revision_id))
        sender = EmailSender(
            master_tree.branch,
            revision_id,
            master_tree.branch.get_config(),
            local_branch=child_tree.branch,
        )
        # We should be using the master repository here, because the child
        # repository doesn't contain the revision.
        self.assertIs(master_tree.branch.repository, sender.repository)



# Copyright (C) 2007-2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for win32utils."""

import os

from .. import osutils, tests, win32utils
from ..win32utils import get_app_path, glob_expand
from . import TestCase, TestCaseInTempDir, TestSkipped, features
from .features import backslashdir_feature

Win32RegistryFeature = features.ModuleAvailableFeature("_winreg")


class TestWin32UtilsGlobExpand(TestCaseInTempDir):
    _test_needs_features: list[features.Feature] = []

    def test_empty_tree(self):
        self.build_tree([])
        self._run_testset(
            [[["a"], ["a"]], [["?"], ["?"]], [["*"], ["*"]], [["a", "a"], ["a", "a"]]]
        )

    def build_ascii_tree(self):
        self.build_tree(
            [
                "a",
                "a1",
                "a2",
                "a11",
                "a.1",
                "b",
                "b1",
                "b2",
                "b3",
                "c/",
                "c/c1",
                "c/c2",
                "d/",
                "d/d1",
                "d/d2",
                "d/e/",
                "d/e/e1",
            ]
        )

    def build_unicode_tree(self):
        self.requireFeature(features.UnicodeFilenameFeature)
        self.build_tree(["\u1234", "\u1234\u1234", "\u1235/", "\u1235/\u1235"])

    def test_tree_ascii(self):
        """Checks the glob expansion and path separation char
        normalization.
        """
        self.build_ascii_tree()
        self._run_testset(
            [
                # no wildcards
                [["a"], ["a"]],
                [["a", "a"], ["a", "a"]],
                [["d"], ["d"]],
                [["d/"], ["d/"]],
                # wildcards
                [["a*"], ["a", "a1", "a2", "a11", "a.1"]],
                [["?"], ["a", "b", "c", "d"]],
                [["a?"], ["a1", "a2"]],
                [["a??"], ["a11", "a.1"]],
                [["b[1-2]"], ["b1", "b2"]],
                [["d/*"], ["d/d1", "d/d2", "d/e"]],
                [["?/*"], ["c/c1", "c/c2", "d/d1", "d/d2", "d/e"]],
                [["*/*"], ["c/c1", "c/c2", "d/d1", "d/d2", "d/e"]],
                [["*/"], ["c/", "d/"]],
            ]
        )

    def test_backslash_globbing(self):
        self.requireFeature(backslashdir_feature)
        self.build_ascii_tree()
        self._run_testset(
            [
                [["d\\"], ["d/"]],
                [["d\\*"], ["d/d1", "d/d2", "d/e"]],
                [["?\\*"], ["c/c1", "c/c2", "d/d1", "d/d2", "d/e"]],
                [["*\\*"], ["c/c1", "c/c2", "d/d1", "d/d2", "d/e"]],
                [["*\\"], ["c/", "d/"]],
            ]
        )

    def test_case_insensitive_globbing(self):
        if os.path.normcase("AbC") == "AbC":
            self.skipTest("Test requires case insensitive globbing function")
        self.build_ascii_tree()
        self._run_testset(
            [
                [["A"], ["A"]],
                [["A?"], ["a1", "a2"]],
            ]
        )

    def test_tree_unicode(self):
        """Checks behaviour with non-ascii filenames."""
        self.build_unicode_tree()
        self._run_testset(
            [
                # no wildcards
                [["\u1234"], ["\u1234"]],
                [["\u1235"], ["\u1235"]],
                [["\u1235/"], ["\u1235/"]],
                [["\u1235/\u1235"], ["\u1235/\u1235"]],
                # wildcards
                [["?"], ["\u1234", "\u1235"]],
                [["*"], ["\u1234", "\u1234\u1234", "\u1235"]],
                [["\u1234*"], ["\u1234", "\u1234\u1234"]],
                [["\u1235/?"], ["\u1235/\u1235"]],
                [["\u1235/*"], ["\u1235/\u1235"]],
                [["?/"], ["\u1235/"]],
                [["*/"], ["\u1235/"]],
                [["?/?"], ["\u1235/\u1235"]],
                [["*/*"], ["\u1235/\u1235"]],
            ]
        )

    def test_unicode_backslashes(self):
        self.requireFeature(backslashdir_feature)
        self.build_unicode_tree()
        self._run_testset(
            [
                # no wildcards
                [["\u1235\\"], ["\u1235/"]],
                [["\u1235\\\u1235"], ["\u1235/\u1235"]],
                [["\u1235\\?"], ["\u1235/\u1235"]],
                [["\u1235\\*"], ["\u1235/\u1235"]],
                [["?\\"], ["\u1235/"]],
                [["*\\"], ["\u1235/"]],
                [["?\\?"], ["\u1235/\u1235"]],
                [["*\\*"], ["\u1235/\u1235"]],
            ]
        )

    def _run_testset(self, testset):
        for pattern, expected in testset:
            result = glob_expand(pattern)
            expected.sort()
            result.sort()
            self.assertEqual(expected, result, "pattern {}".format(pattern))


class TestAppPaths(TestCase):
    _test_needs_features = [Win32RegistryFeature]

    def test_iexplore(self):
        # typical windows users should have IE installed
        for a in ("iexplore", "iexplore.exe"):
            p = get_app_path(a)
            d, b = os.path.split(p)
            self.assertEqual("iexplore.exe", b.lower())
            self.assertNotEqual("", d)

    def test_wordpad(self):
        # typical windows users should have wordpad in the system
        # but there is problem: its path has the format REG_EXPAND_SZ
        # so naive attempt to get the path is not working
        self.requireFeature(Win32ApiFeature)
        for a in ("wordpad", "wordpad.exe"):
            p = get_app_path(a)
            d, b = os.path.split(p)
            self.assertEqual("wordpad.exe", b.lower())
            self.assertNotEqual("", d)

    def test_not_existing(self):
        p = get_app_path("not-existing")
        self.assertEqual("not-existing", p)


class TestLocations(TestCase):
    _test_needs_features = [features.win32_feature]

    def assertPathsEqual(self, p1, p2):
        # TODO: The env var values in particular might return the "short"
        # version (ie, "C:\DOCUME~1\...").  Its even possible the returned
        # values will differ only by case - handle these situations as we
        # come across them.
        self.assertEqual(p1, p2)

    def test_appdata_not_using_environment(self):
        # Test that we aren't falling back to the environment
        first = win32utils.get_appdata_location()
        self.overrideEnv("APPDATA", None)
        self.assertPathsEqual(first, win32utils.get_appdata_location())

    def test_appdata_matches_environment(self):
        # Typically the APPDATA environment variable will match
        # get_appdata_location
        # XXX - See bug 262874, which asserts the correct encoding is 'mbcs',
        encoding = osutils.get_user_encoding()
        env_val = os.environ.get("APPDATA", None)
        if not env_val:
            raise TestSkipped("No APPDATA environment variable exists")
        self.assertPathsEqual(
            win32utils.get_appdata_location(), env_val.decode(encoding)
        )

    def test_local_appdata_not_using_environment(self):
        # Test that we aren't falling back to the environment
        first = win32utils.get_local_appdata_location()
        self.overrideEnv("LOCALAPPDATA", None)
        self.assertPathsEqual(first, win32utils.get_local_appdata_location())

    def test_local_appdata_matches_environment(self):
        # LOCALAPPDATA typically only exists on Vista, so we only attempt to
        # compare when it exists.
        lad = win32utils.get_local_appdata_location()
        env = os.environ.get("LOCALAPPDATA")
        if env:
            # XXX - See bug 262874, which asserts the correct encoding is
            # 'mbcs'
            encoding = osutils.get_user_encoding()
            self.assertPathsEqual(lad, env.decode(encoding))


class TestSetHidden(TestCaseInTempDir):
    _test_needs_features = [features.win32_feature]

    def test_unicode_dir(self):
        # we should handle unicode paths without errors
        self.requireFeature(features.UnicodeFilenameFeature)
        os.mkdir("\u1234")
        win32utils.set_file_attr_hidden("\u1234")

    def test_dot_bzr_in_unicode_dir(self):
        # we should not raise traceback if we try to set hidden attribute
        # on .bzr directory below unicode path
        self.requireFeature(features.UnicodeFilenameFeature)
        os.makedirs("\u1234\\.bzr")
        path = osutils.abspath("\u1234\\.bzr")
        win32utils.set_file_attr_hidden(path)


class Test_CommandLineToArgv(tests.TestCaseInTempDir):
    def assertCommandLine(self, expected, line, argv=None, single_quotes_allowed=False):
        # Strictly speaking we should respect parameter order versus glob
        # expansions, but it's not really worth the effort here
        if argv is None:
            argv = [line]
        argv = win32utils._command_line_to_argv(
            line, argv, single_quotes_allowed=single_quotes_allowed
        )
        self.assertEqual(expected, sorted(argv))

    def test_glob_paths(self):
        self.build_tree(["a/", "a/b.c", "a/c.c", "a/c.h"])
        self.assertCommandLine(["a/b.c", "a/c.c"], "a/*.c")
        self.build_tree(["b/", "b/b.c", "b/d.c", "b/d.h"])
        self.assertCommandLine(["a/b.c", "b/b.c"], "*/b.c")
        self.assertCommandLine(["a/b.c", "a/c.c", "b/b.c", "b/d.c"], "*/*.c")
        # Bash style, just pass through the argument if nothing matches
        self.assertCommandLine(["*/*.qqq"], "*/*.qqq")

    def test_quoted_globs(self):
        self.build_tree(["a/", "a/b.c", "a/c.c", "a/c.h"])
        self.assertCommandLine(["a/*.c"], '"a/*.c"')
        self.assertCommandLine(["'a/*.c'"], "'a/*.c'")
        self.assertCommandLine(["a/*.c"], "'a/*.c'", single_quotes_allowed=True)

    def test_slashes_changed(self):
        # Quoting doesn't change the supplied args
        self.assertCommandLine(["a\\*.c"], '"a\\*.c"')
        self.assertCommandLine(["a\\*.c"], "'a\\*.c'", single_quotes_allowed=True)
        # Expands the glob, but nothing matches, swaps slashes
        self.assertCommandLine(["a/*.c"], "a\\*.c")
        self.assertCommandLine(["a/?.c"], "a\\?.c")
        # No glob, doesn't touch slashes
        self.assertCommandLine(["a\\foo.c"], "a\\foo.c")

    def test_single_quote_support(self):
        self.assertCommandLine(
            ["add", "let's-do-it.txt"],
            "add let's-do-it.txt",
            ["add", "let's-do-it.txt"],
        )
        self.expectFailure(
            "Using single quotes breaks trimming from argv",
            self.assertCommandLine,
            ["add", "lets do it.txt"],
            "add 'lets do it.txt'",
            ["add", "'lets", "do", "it.txt'"],
            single_quotes_allowed=True,
        )

    def test_case_insensitive_globs(self):
        if os.path.normcase("AbC") == "AbC":
            self.skipTest("Test requires case insensitive globbing function")
        self.build_tree(["a/", "a/b.c", "a/c.c", "a/c.h"])
        self.assertCommandLine(["A/b.c"], "A/B*")

    def test_backslashes(self):
        self.requireFeature(backslashdir_feature)
        self.build_tree(["a/", "a/b.c", "a/c.c", "a/c.h"])
        self.assertCommandLine(["a/b.c"], "a\\b*")

    def test_with_pdb(self):
        """Check stripping Python arguments before bzr script per lp:587868."""
        self.assertCommandLine(["rocks"], "-m pdb rocks", ["rocks"])
        self.build_tree(["d/", "d/f1", "d/f2"])
        self.assertCommandLine(["rm", "x*"], "-m pdb rm x*", ["rm", "x*"])
        self.assertCommandLine(
            ["add", "d/f1", "d/f2"], "-m pdb add d/*", ["add", "d/*"]
        )



#!/usr/bin/python3

import ast

from pyflakes.checker import Checker, ImportationFrom


# Do some monkey patching..
def CALL(self, node):
    if isinstance(node.func, ast.Name) and node.func.id == "lazy_import":
        self.LAZY_IMPORT(node)
    elif (
        isinstance(node.func, ast.Attribute)
        and node.func.attr == "lazy_import"
        and node.func.value.id == "lazy_import"
    ):
        self.LAZY_IMPORT(node)
    return self.handleChildren(node)


Checker.CALL = CALL


def LAZY_IMPORT(self, node):
    from breezy.lazy_import import ImportProcessor

    processor = ImportProcessor()
    if not isinstance(node.args[1], ast.Str):
        # Not sure how to deal with this..
        return
    import_text = node.args[1].s
    scope = {}
    processor.lazy_import(scope, import_text)
    for name, (path, _sub, scope) in processor.imports.items():
        importation = ImportationFrom(name, node, ".".join(path), scope)
        self.addBinding(node, importation)


Checker.LAZY_IMPORT = LAZY_IMPORT


if __name__ == "__main__":
    from pyflakes import __main__

    __main__.main()



# Copyright (C) 2006, 2007, 2009-2012, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for brz help."""

from breezy import config, i18n, tests
from breezy.tests.test_i18n import ZzzTranslations


class TestHelp(tests.TestCaseWithTransport):
    def test_help_basic(self):
        for cmd in ["--help", "help", "-h", "-?"]:
            output = self.run_bzr(cmd)[0]
            line1 = output.split("\n")[0]
            if not line1.startswith("Breezy"):
                self.fail("bad output from brz {}:\n{!r}".format(cmd, output))
        # see https://launchpad.net/products/bzr/+bug/35940, -h doesn't work

    def test_help_topics(self):
        """Smoketest for 'brz help topics'."""
        out, err = self.run_bzr("help topics")
        self.assertContainsRe(out, "basic")
        self.assertContainsRe(out, "topics")
        self.assertContainsRe(out, "commands")
        self.assertContainsRe(out, "revisionspec")

    def test_help_revisionspec(self):
        """Smoke test for 'brz help revisionspec'."""
        out, err = self.run_bzr("help revisionspec")
        self.assertContainsRe(out, "revno:")
        self.assertContainsRe(out, "date:")
        self.assertContainsRe(out, "revid:")
        self.assertContainsRe(out, "last:")
        self.assertContainsRe(out, "before:")
        self.assertContainsRe(out, "ancestor:")
        self.assertContainsRe(out, "branch:")

    def test_help_checkouts(self):
        """Smoke test for 'brz help checkouts'."""
        out, err = self.run_bzr("help checkouts")
        self.assertContainsRe(out, "checkout")
        self.assertContainsRe(out, "lightweight")

    def test_help_urlspec(self):
        """Smoke test for 'brz help urlspec'."""
        out, err = self.run_bzr("help urlspec")
        self.assertContainsRe(out, "bzr://")
        self.assertContainsRe(out, "bzr\\+ssh://")
        self.assertContainsRe(out, "file://")
        self.assertContainsRe(out, "http://")
        self.assertContainsRe(out, "https://")
        self.assertContainsRe(out, "sftp://")

    def test_help_repositories(self):
        """Smoke test for 'brz help repositories'."""
        out, err = self.run_bzr("help repositories")
        from breezy.help_topics import _repositories, help_as_plain_text

        expected = help_as_plain_text(_repositories)
        self.assertEqual(expected, out)

    def test_help_working_trees(self):
        """Smoke test for 'brz help working-trees'."""
        out, err = self.run_bzr("help working-trees")
        from breezy.help_topics import _working_trees, help_as_plain_text

        expected = help_as_plain_text(_working_trees)
        self.assertEqual(expected, out)

    def test_help_status_flags(self):
        """Smoke test for 'brz help status-flags'."""
        out, err = self.run_bzr("help status-flags")
        from breezy.help_topics import _status_flags, help_as_plain_text

        expected = help_as_plain_text(_status_flags)
        self.assertEqual(expected, out)

    def test_help_commands(self):
        dash_help = self.run_bzr("--help commands")[0]
        commands = self.run_bzr("help commands")[0]
        self.run_bzr("help hidden-commands")[0]
        long_help = self.run_bzr("help --long")[0]
        qmark_long = self.run_bzr("? --long")[0]
        qmark_cmds = self.run_bzr("? commands")[0]
        self.assertEqual(dash_help, commands)
        self.assertEqual(dash_help, long_help)
        self.assertEqual(dash_help, qmark_long)
        self.assertEqual(dash_help, qmark_cmds)

    def test_help_width_zero(self):
        self.overrideEnv("BRZ_COLUMNS", "0")
        self.run_bzr("help commands")

    def test_hidden(self):
        help_commands = self.run_bzr("help commands")[0]
        help_hidden = self.run_bzr("help hidden-commands")[0]

        def extract_cmd_names(help_output):
            # keep only the command names to avoid matching on help text (there
            # is a high risk to fail a test when a plugin get installed
            # otherwise)
            cmds = []
            for line in help_output.split("\n"):
                if line.startswith(" "):
                    continue  # help on more than one line
                cmd = line.split(" ")[0]
                if line:
                    cmds.append(cmd)
            return cmds

        commands = extract_cmd_names(help_commands)
        hidden = extract_cmd_names(help_hidden)
        self.assertTrue("commit" in commands)
        self.assertTrue("commit" not in hidden)
        self.assertTrue("rocks" in hidden)
        self.assertTrue("rocks" not in commands)

    def test_help_detail(self):
        dash_h = self.run_bzr("diff -h")[0]
        help_x = self.run_bzr("help diff")[0]
        self.assertEqual(dash_h, help_x)
        self.assertContainsRe(help_x, "Purpose:")
        self.assertContainsRe(help_x, "Usage:")
        self.assertContainsRe(help_x, "Options:")
        self.assertContainsRe(help_x, "Description:")
        self.assertContainsRe(help_x, "Examples:")
        self.assertContainsRe(help_x, "See also:")
        self.assertContainsRe(help_x, "Aliases:")

    def test_help_usage(self):
        usage = self.run_bzr("diff --usage")[0]
        self.assertContainsRe(usage, "Purpose:")
        self.assertContainsRe(usage, "Usage:")
        self.assertContainsRe(usage, "Options:")
        self.assertNotContainsRe(usage, "Description:")
        self.assertNotContainsRe(usage, "Examples:")
        self.assertContainsRe(usage, "See also:")
        self.assertContainsRe(usage, "Aliases:")

    def test_help_help(self):
        help = self.run_bzr("help help")[0]
        qmark = self.run_bzr("? ?")[0]
        self.assertEqual(help, qmark)
        for line in help.split("\n"):
            if "--long" in line:
                self.assertContainsRe(line, r"Show help on all commands\.")

    def test_help_with_aliases(self):
        original = self.run_bzr("help cat")[0]

        config.GlobalConfig.from_string(
            """[ALIASES]
c=cat
cat=cat
""",
            save=True,
        )

        expected = original + "'brz cat' is an alias for 'brz cat'.\n"
        self.assertEqual(expected, self.run_bzr("help cat")[0])

        self.assertEqual(
            "'brz c' is an alias for 'brz cat'.\n", self.run_bzr("help c")[0]
        )


class TestTranslatedHelp(tests.TestCaseWithTransport):
    """Tests for display of translated help topics."""

    def setUp(self):
        super().setUp()
        self.overrideAttr(i18n, "_translations", ZzzTranslations())

    def test_help_command_utf8(self):
        out, err = self.run_bzr_raw(["help", "push"], encoding="utf-8")
        self.assertContainsRe(out, b"zz\xc3\xa5{{:See also:")

    def test_help_switch_utf8(self):
        out, err = self.run_bzr_raw(["push", "--help"], encoding="utf-8")
        self.assertContainsRe(out, b"zz\xc3\xa5{{:See also:")

    def test_help_command_ascii(self):
        out, err = self.run_bzr_raw(["help", "push"], encoding="ascii")
        self.assertContainsRe(out, b"zz\\?{{:See also:")

    def test_help_switch_ascii(self):
        out, err = self.run_bzr_raw(["push", "--help"], encoding="ascii")
        self.assertContainsRe(out, b"zz\\?{{:See also:")



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test our ability to build up test repositories."""

from io import BytesIO

from dulwich.repo import Repo as GitRepo

from .. import tests


class TestGitBranchBuilder(tests.TestCase):
    def test__create_blob(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        self.assertEqual(1, builder._create_blob(b"foo\nbar\n"))
        self.assertEqualDiff(b"blob\nmark :1\ndata 8\nfoo\nbar\n\n", stream.getvalue())

    def test_set_file(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_file("foobar", b"foo\nbar\n", False)
        self.assertEqualDiff(b"blob\nmark :1\ndata 8\nfoo\nbar\n\n", stream.getvalue())
        self.assertEqual([b"M 100644 :1 foobar\n"], builder.commit_info)

    def test_set_file_unicode(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_file("f\xb5/bar", b"contents\nbar\n", False)
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nbar\n\n", stream.getvalue()
        )
        self.assertEqual([b"M 100644 :1 f\xc2\xb5/bar\n"], builder.commit_info)

    def test_set_file_newline(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_file("foo\nbar", b"contents\nbar\n", False)
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nbar\n\n", stream.getvalue()
        )
        self.assertEqual([b'M 100644 :1 "foo\\nbar"\n'], builder.commit_info)

    def test_set_file_executable(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_file("f\xb5/bar", b"contents\nbar\n", True)
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nbar\n\n", stream.getvalue()
        )
        self.assertEqual([b"M 100755 :1 f\xc2\xb5/bar\n"], builder.commit_info)

    def test_set_symlink(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_symlink("f\xb5/bar", b"link/contents")
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\nlink/contents\n", stream.getvalue()
        )
        self.assertEqual([b"M 120000 :1 f\xc2\xb5/bar\n"], builder.commit_info)

    def test_set_symlink_newline(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.set_symlink("foo\nbar", "link/contents")
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\nlink/contents\n", stream.getvalue()
        )
        self.assertEqual([b'M 120000 :1 "foo\\nbar"\n'], builder.commit_info)

    def test_delete_entry(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.delete_entry("path/to/f\xb5")
        self.assertEqual([b"D path/to/f\xc2\xb5\n"], builder.commit_info)

    def test_delete_entry_newline(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.delete_entry("path/to/foo\nbar")
        self.assertEqual([b'D "path/to/foo\\nbar"\n'], builder.commit_info)

    def test_encode_path(self):
        encode = tests.GitBranchBuilder._encode_path
        # Unicode is encoded to utf-8
        self.assertEqual(encode("f\xb5"), b"f\xc2\xb5")
        # The name must be quoted if it starts by a double quote or contains a
        # newline.
        self.assertEqual(encode('"foo'), b'"\\"foo"')
        self.assertEqual(encode("fo\no"), b'"fo\\no"')
        # When the name is quoted, all backslash and quote chars must be
        # escaped.
        self.assertEqual(encode("fo\\o\nbar"), b'"fo\\\\o\\nbar"')
        self.assertEqual(encode('fo"o"\nbar'), b'"fo\\"o\\"\\nbar"')
        # Other control chars, such as \r, need not be escaped.
        self.assertEqual(encode("foo\r\nbar"), b'"foo\r\\nbar"')

    def test_add_and_commit(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)

        builder.set_file("f\xb5/bar", b"contents\nbar\n", False)
        self.assertEqual(
            b"2",
            builder.commit(
                b"Joe Foo <joe@foo.com>",
                "committing f\xb5/bar",
                timestamp=1194586400,
                timezone=b"+0100",
            ),
        )
        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nbar\n\n"
            b"commit refs/heads/master\n"
            b"mark :2\n"
            b"committer Joe Foo <joe@foo.com> 1194586400 +0100\n"
            b"data 18\n"
            b"committing f\xc2\xb5/bar"
            b"\n"
            b"M 100644 :1 f\xc2\xb5/bar\n"
            b"\n",
            stream.getvalue(),
        )

    def test_commit_base(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)

        builder.set_file("foo", b"contents\nfoo\n", False)
        r1 = builder.commit(b"Joe Foo <joe@foo.com>", "first", timestamp=1194586400)
        builder.commit(b"Joe Foo <joe@foo.com>", "second", timestamp=1194586405)
        builder.commit(b"Joe Foo <joe@foo.com>", "third", timestamp=1194586410, base=r1)

        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nfoo\n\n"
            b"commit refs/heads/master\n"
            b"mark :2\n"
            b"committer Joe Foo <joe@foo.com> 1194586400 +0000\n"
            b"data 5\n"
            b"first"
            b"\n"
            b"M 100644 :1 foo\n"
            b"\n"
            b"commit refs/heads/master\n"
            b"mark :3\n"
            b"committer Joe Foo <joe@foo.com> 1194586405 +0000\n"
            b"data 6\n"
            b"second"
            b"\n"
            b"\n"
            b"commit refs/heads/master\n"
            b"mark :4\n"
            b"committer Joe Foo <joe@foo.com> 1194586410 +0000\n"
            b"data 5\n"
            b"third"
            b"\n"
            b"from :2\n"
            b"\n",
            stream.getvalue(),
        )

    def test_commit_merge(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)

        builder.set_file("foo", b"contents\nfoo\n", False)
        r1 = builder.commit(b"Joe Foo <joe@foo.com>", "first", timestamp=1194586400)
        r2 = builder.commit(b"Joe Foo <joe@foo.com>", "second", timestamp=1194586405)
        builder.commit(b"Joe Foo <joe@foo.com>", "third", timestamp=1194586410, base=r1)
        builder.commit(
            b"Joe Foo <joe@foo.com>", "Merge", timestamp=1194586415, merge=[r2]
        )

        self.assertEqualDiff(
            b"blob\nmark :1\ndata 13\ncontents\nfoo\n\n"
            b"commit refs/heads/master\n"
            b"mark :2\n"
            b"committer Joe Foo <joe@foo.com> 1194586400 +0000\n"
            b"data 5\n"
            b"first"
            b"\n"
            b"M 100644 :1 foo\n"
            b"\n"
            b"commit refs/heads/master\n"
            b"mark :3\n"
            b"committer Joe Foo <joe@foo.com> 1194586405 +0000\n"
            b"data 6\n"
            b"second"
            b"\n"
            b"\n"
            b"commit refs/heads/master\n"
            b"mark :4\n"
            b"committer Joe Foo <joe@foo.com> 1194586410 +0000\n"
            b"data 5\n"
            b"third"
            b"\n"
            b"from :2\n"
            b"\n"
            b"commit refs/heads/master\n"
            b"mark :5\n"
            b"committer Joe Foo <joe@foo.com> 1194586415 +0000\n"
            b"data 5\n"
            b"Merge"
            b"\n"
            b"merge :3\n"
            b"\n",
            stream.getvalue(),
        )

    def test_auto_timestamp(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.commit(b"Joe Foo <joe@foo.com>", "message")
        self.assertContainsRe(
            stream.getvalue(), rb"committer Joe Foo <joe@foo\.com> \d+ \+0000"
        )

    def test_reset(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.reset()
        self.assertEqualDiff(b"reset refs/heads/master\n\n", stream.getvalue())

    def test_reset_named_ref(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.reset(b"refs/heads/branch")
        self.assertEqualDiff(b"reset refs/heads/branch\n\n", stream.getvalue())

    def test_reset_revision(self):
        stream = BytesIO()
        builder = tests.GitBranchBuilder(stream)
        builder.reset(mark=b"123")
        self.assertEqualDiff(
            b"reset refs/heads/master\nfrom :123\n\n", stream.getvalue()
        )


class TestGitBranchBuilderReal(tests.TestCaseInTempDir):
    def test_create_real_branch(self):
        GitRepo.init(".")

        builder = tests.GitBranchBuilder()
        builder.set_file("foo", b"contents\nfoo\n", False)
        builder.commit(b"Joe Foo <joe@foo.com>", "first", timestamp=1194586400)
        mapping = builder.finish()
        self.assertEqual(
            {
                b"1": b"44411e8e9202177dd19b6599d7a7991059fa3cb4",
                b"2": b"b0b62e674f67306fddcf72fa888c3b56df100d64",
            },
            mapping,
        )



#
# Copyright (C) 2007, 2009, 2011 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# This code is from bzr-explorer and modified for bzr.

"""build_mo command for setup.py."""

import os
import re
from distutils import log  # type: ignore
from distutils.core import Command
from distutils.dep_util import newer
from distutils.spawn import find_executable
from typing import Optional


class build_mo(Command):
    """Subcommand of build command: build_mo."""

    description = "compile po files to mo files"

    # List of options:
    #   - long name,
    #   - short name (None if no short name),
    #   - help string.
    user_options = [
        ("build-dir=", "d", "Directory to build locale files"),
        ("output-base=", "o", "mo-files base name"),
        ("source-dir=", None, "Directory with sources po files"),
        ("force", "f", "Force creation of mo files"),
        ("lang=", None, "Comma-separated list of languages to process"),
    ]

    boolean_options = ["force"]
    source_dir: Optional[str]
    build_dir: Optional[str]
    output_base: Optional[str]
    lang: Optional[list[str]]

    def initialize_options(self) -> None:
        self.build_dir = None
        self.output_base = None
        self.source_dir = None  # type: ignore
        self.force = None
        self.lang = None

    def finalize_options(self):
        self.set_undefined_options("build", ("force", "force"))
        self.prj_name = self.distribution.get_name()  # type: ignore
        if self.build_dir is None:
            self.build_dir = "breezy/locale"
        if not self.output_base:
            self.output_base = self.prj_name or "messages"
        if self.source_dir is None:
            self.source_dir = "po"
        if self.lang is None:
            re_po = re.compile(r"^([a-zA-Z_]+)\.po$")
            self.lang = []
            for i in os.listdir(self.source_dir):
                mo = re_po.match(i)
                if mo:
                    self.lang.append(mo.group(1))
        else:
            self.lang = [i.strip() for i in self.lang.split(",") if i.strip()]

    def run(self):
        """Run msgfmt for each language."""
        if not self.lang:
            return

        if find_executable("msgfmt") is None:
            log.warn("GNU gettext msgfmt utility not found!")
            log.warn("Skip compiling po files.")
            return

        if "en" in self.lang:
            if find_executable("msginit") is None:
                log.warn("GNU gettext msginit utility not found!")
                log.warn("Skip creating English PO file.")
            else:
                log.info("Creating English PO file...")
                pot = (self.prj_name or "messages") + ".pot"
                en_po = "en.po"
                self.spawn(
                    [
                        "msginit",
                        "--no-translator",
                        "-l",
                        "en",
                        "-i",
                        os.path.join(self.source_dir, pot),
                        "-o",
                        os.path.join(self.source_dir, en_po),
                    ]
                )

        basename = self.output_base
        if not basename.endswith(".mo"):
            basename += ".mo"

        for lang in self.lang:
            po = os.path.join("po", lang + ".po")
            if not os.path.isfile(po):
                po = os.path.join("po", lang + ".po")
            dir_ = os.path.join(self.build_dir, lang, "LC_MESSAGES")
            self.mkpath(dir_)
            mo = os.path.join(dir_, basename)
            if self.force or newer(po, mo):
                log.info("Compile: {} -> {}".format(po, mo))
                self.spawn(["msgfmt", "-o", mo, po])
